from __future__ import absolute_import

from unittest import TestCase

from simulation.location import Location


class TestLocation(TestCase):
    def test_equal(self):
        loc_1 = Location(3, 3)
        loc_2 = Location(3, 3)
        self.assertEqual(loc_1, loc_2)
        self.assertFalse(loc_1 != loc_2)

    def test_x_not_equal(self):
        loc_1 = Location(3, 3)
        loc_2 = Location(4, 3)
        self.assertNotEqual(loc_1, loc_2)
        self.assertFalse(loc_1 == loc_2)

    def test_y_not_equal(self):
        loc_1 = Location(4, 4)
        loc_2 = Location(4, 3)
        self.assertNotEqual(loc_1, loc_2)
        self.assertFalse(loc_1 == loc_2)

    def test_add(self):
        loc_1 = Location(1, 2)
        loc_2 = Location(3, 4)
        expected = Location(4, 6)
        self.assertEqual(loc_1 + loc_2, expected)

    def test_sub(self):
        loc_1 = Location(1, 2)
        loc_2 = Location(3, 4)
        expected = Location(-2, -2)
        self.assertEqual(loc_1 - loc_2, expected)

        loc_2 = Location(0, 0)
        expected = Location(1, 2)
        self.assertEqual(loc_1 - loc_2, expected)

    def test_hash_equal(self):
        loc_1 = Location(3, 3)
        loc_2 = Location(3, 3)
        self.assertEqual(hash(loc_1), hash(loc_2))

    def test_location_raises_exception_with_floats(self):
        self.assertRaises(TypeError, Location, 3.2, 3)
        self.assertRaises(TypeError, Location, 2, 2.2)
        self.assertRaises(TypeError, Location, 2.5, 2.5)
        self.assertRaises(TypeError, Location, 2.0, 1)



from collections import defaultdict

from simulation.cell import Cell
from simulation.game_logic import SpawnLocationFinder
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location
from simulation.world_map import WorldMap


class MockPickup(object):
    def __init__(self, name="", cell=None, target=None):
        self.applied_to = target
        self.name = name
        self.cell = None

    def apply(self):
        if self.cell:
            self.cell.interactable = None

    def conditions_met(self, world_map):
        return True

    def serialize(self):
        return {"name": self.name}


class MockCell(Cell):
    def __init__(
        self,
        location=Location(0, 0),
        habitable=True,
        avatar=None,
        interactable=None,
        name=None,
        actions=[],
    ):
        self.location = location
        self.habitable = habitable
        self.avatar = avatar
        self.interactable = interactable
        self.name = name
        self.actions = actions
        self.partially_fogged = False

    def __eq__(self, other):
        return self is other


class InfiniteMap(WorldMap):
    def __init__(self):
        self._spawn_location_finder = SpawnLocationFinder(self)
        self._cell_cache = {}
        [self.get_cell(Location(x, y)) for x in range(5) for y in range(5)]
        self.updates = 0
        self.num_avatars = None
        self.settings = defaultdict(lambda: 0)

    def is_on_map(self, target_location):
        self.get_cell(target_location)
        return True

    def all_cells(self):
        return (cell for cell in self._cell_cache.values())

    def get_cell(self, location):
        return self._cell_cache.setdefault(location, Cell(location))

    def update(self, num_avatars):
        self.updates += 1
        self.num_avatars = num_avatars

    @property
    def num_rows(self):
        return float("inf")

    @property
    def num_cols(self):
        return float("inf")


class EmptyMap(WorldMap):
    def __init__(self):
        pass

    def get_random_spawn_location(self):
        return Location(10, 10)

    def can_move_to(self, target_location):
        return False

    def all_cells(self):
        return iter(())

    def get_cell(self, location):
        return Cell(location)


class ScoreOnOddColumnsMap(InfiniteMap):
    def get_cell(self, location):
        default_cell = Cell(location)
        if location.x % 2 == 1:
            default_cell.interactable = ScoreLocation(default_cell)
        return self._cell_cache.setdefault(location, default_cell)


class AvatarMap(WorldMap):
    def __init__(self, avatar):
        self._avatar = avatar
        self._cell_cache = {}

    def get_cell(self, location):
        if location not in self._cell_cache:
            cell = Cell(location)
            cell.avatar = self._avatar
            self._cell_cache[location] = cell
        return self._cell_cache[location]

    def get_random_spawn_location(self):
        return Location(10, 10)


class PickupMap(WorldMap):
    def __init__(self, pickup):
        self._pickup = pickup

    def get_cell(self, location):
        cell = Cell(location)
        cell.interactable = self._pickup
        return cell



from json import dumps

import pytest

import mock
from simulation.avatar.avatar_manager import AvatarManager
from simulation.game_runner import GameRunner
from simulation.game_state import GameState

from .maps import InfiniteMap
from .mock_communicator import MockCommunicator
from .mock_worker_manager import MockWorkerManager


class RequestMock(object):
    def __init__(self, num_users):
        self.value = self._generate_response(num_users)
        self.urls_requested = []

    def _generate_response(self, num_users):
        return {
            "main_avatar": None,
            "users": [{"id": i, "code": "code for %s" % i} for i in range(num_users)],
        }

    def change_code(self, id, new_code):
        users = self.value["users"]
        for i in range(len(users)):
            if users[i]["id"] == id:
                users[i]["code"] = new_code

    def __call__(self, url, request):
        self.urls_requested.append(url.geturl())
        return dumps(self.value)


@pytest.fixture
def game_runner():
    async def mock_callback():
        pass

    game_state = GameState(InfiniteMap(), AvatarManager())
    game_runner = GameRunner(
        game_state_generator=lambda avatar_manager: game_state,
        port="0000",
        communicator=MockCommunicator(),
        worker_manager_class=MockWorkerManager,
    )

    game_runner.set_end_turn_callback(mock_callback)
    return game_runner


@pytest.mark.asyncio
async def test_correct_url(game_runner):
    game_runner.communicator.get_game_metadata = mock.MagicMock()
    await game_runner.update()
    # noinspection PyUnresolvedReferences
    game_runner.communicator.get_game_metadata.assert_called_once()


@pytest.mark.asyncio
async def test_workers_and_avatars_added(game_runner):
    game_runner.communicator.data = RequestMock(3).value
    await game_runner.update()

    assert len(game_runner.worker_manager.final_workers) == 3
    for i in range(3):
        assert i in game_runner.game_state.avatar_manager.avatars_by_id
        assert i in game_runner.worker_manager.final_workers
        assert game_runner.worker_manager.get_code(i) == "code for %s" % i


@pytest.mark.asyncio
async def test_changed_code(game_runner):
    game_runner.communicator.data = RequestMock(4).value
    await game_runner.update()
    game_runner.communicator.change_code(0, "changed 0")
    game_runner.communicator.change_code(2, "changed 2")
    await game_runner.update()

    for i in range(4):
        assert i in game_runner.worker_manager.final_workers
        assert i in game_runner.game_state.avatar_manager.avatars_by_id

    for i in (1, 3):
        assert game_runner.worker_manager.get_code(i) in "code for %s" % i

    for i in (0, 2):
        assert i in game_runner.worker_manager.updated_workers
        assert game_runner.worker_manager.get_code(i) in "changed %s" % i


@pytest.mark.asyncio
async def test_logs_cleared_at_each_update(game_runner):
    game_runner.communicator.data = RequestMock(3).value
    await game_runner.update_workers()
    first_worker = game_runner.worker_manager.player_id_to_worker[0]
    first_worker.log = "test logs"

    game_runner.worker_manager.clear_logs()

    assert first_worker.log is None


@pytest.mark.asyncio
async def test_remove_avatars(game_runner):
    game_runner.communicator.data = RequestMock(3).value
    await game_runner.update()
    del game_runner.communicator.data["users"][1]
    await game_runner.update()

    for i in range(3):
        if i == 1:
            assert i not in game_runner.worker_manager.final_workers
            assert i not in game_runner.game_state.avatar_manager.avatars_by_id
        else:
            assert i in game_runner.worker_manager.final_workers
            assert i in game_runner.game_state.avatar_manager.avatars_by_id


@pytest.mark.asyncio
async def test_turn_increment(game_runner):
    game_runner.communicator.data = RequestMock(3).value
    assert game_runner.game_state.turn_count == 0
    await game_runner.update()
    assert game_runner.game_state.turn_count == 1



class Direction:
    def __init__(self, x, y):
        if abs(x) not in [0, 1]:
            raise ValueError
        if abs(y) not in [0, 1]:
            raise ValueError
        if abs(x) + abs(y) not in [0, 1]:
            raise ValueError
        self.x = x
        self.y = y

    @property
    def dict(self):
        return {"x": self.x, "y": self.y}

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def cardinal(self):
        if self == NORTH:
            return "north"
        elif self == EAST:
            return "east"
        elif self == WEST:
            return "west"
        elif self == SOUTH:
            return "south"

    def __repr__(self):
        return "Direction(x={}, y={})".format(self.x, self.y)


NORTH = Direction(0, 1)
EAST = Direction(1, 0)
SOUTH = Direction(0, -1)
WEST = Direction(-1, 0)

ALL_DIRECTIONS = (NORTH, EAST, SOUTH, WEST)



import os
import os
import random
import string
from unittest import mock

import pytest
import socketio

import service
from simulation.game_runner import GameRunner
from .test_simulation.mock_communicator import MockCommunicator
from .test_simulation.mock_worker_manager import MockWorkerManager

TIME_TO_PROCESS_SOME_EVENT_LOOP = 0.1


class MockGameState(object):
    turn_count = 0

    def serialize(self):
        return {"foo": "bar"}


@pytest.fixture
def game_id():
    os.environ["GAME_ID"] = "1"
    yield
    del os.environ["GAME_ID"]


@pytest.fixture
def sid():
    return "".join(
        random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits)
        for _ in range(19)
    )


@pytest.fixture
def app():
    return service.setup_application(should_clean_token=False)


@pytest.fixture
def socketio_server(app):
    return service.setup_socketIO_server(app, async_handlers=False)


@pytest.fixture
def game_api(app, socketio_server, game_id):
    game_runner = GameRunner(
        game_state_generator=lambda avatar_manager: MockGameState(),
        communicator=MockCommunicator(),
        port="0000",
        worker_manager_class=MockWorkerManager,
    )
    return service.GameAPI(
        game_state=game_runner.game_state,
        worker_manager=game_runner.worker_manager,
        application=app,
        server=socketio_server,
    )


@pytest.fixture
def client(app, aiohttp_client, loop):
    return loop.run_until_complete(aiohttp_client(app))


async def test_socketio_emit_called_when_worker_ready(
    game_api, socketio_server, client, loop
):
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_game_state_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("game-state", mock_game_state_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker.ready = True

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_game_state_listener.assert_called_once()


async def test_socketio_emit_not_called_if_worker_not_ready(
    game_api, socketio_server, client, loop
):
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_game_state_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("game-state", mock_game_state_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    assert worker.ready == False

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_game_state_listener.assert_not_called()


async def test_send_updates_for_one_user(game_api, client, socketio_server, loop):
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_log_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("log", mock_log_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker.log = "Logs one"

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_log_listener.assert_has_calls(
        [mock.call({"message": "Logs one", "turn_count": 0})]
    )


async def test_no_logs_not_emitted(game_api, client, socketio_server, loop):
    """ If there are no logs for an avatar, no logs should be emitted. """
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_log_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("log", mock_log_listener)

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_log_listener.assert_not_called()


async def test_empty_logs_not_emitted(game_api, client, socketio_server, loop):
    """ If the logs are an empty sting, no logs should be emitted. """
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_log_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("log", mock_log_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker.log = ""

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_log_listener.assert_not_called()


async def test_send_updates_for_multiple_users(game_api, client, socketio_server, loop):
    socketio_client = socketio.AsyncClient(reconnection=False)
    socketio_client2 = socketio.AsyncClient(reconnection=False)
    mock_log_listener = mock.MagicMock()
    mock_log_listener2 = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)
    game_api.worker_manager.add_new_worker(2)

    socketio_client.on("log", mock_log_listener)
    socketio_client2.on("log", mock_log_listener2)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker2 = game_api.worker_manager.player_id_to_worker[2]
    worker.log = "Logs one"
    worker2.log = "Logs two"

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await socketio_client2.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=2&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()
    await socketio_client2.disconnect()

    mock_log_listener.assert_has_calls(
        [mock.call({"message": "Logs one", "turn_count": 0})]
    )
    mock_log_listener2.assert_has_calls(
        [mock.call({"message": "Logs two", "turn_count": 0})]
    )


async def test_send_code_changed_flag(game_api, client, socketio_server, loop):
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_avatar_updated_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("feedback-avatar-updated", mock_avatar_updated_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker.has_code_updated = True

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_avatar_updated_listener.assert_called_once()


async def test_send_false_flag_not_sent(game_api, client, socketio_server, loop):
    socketio_client = socketio.AsyncClient(reconnection=False)
    mock_avatar_updated_listener = mock.MagicMock()

    game_api.worker_manager.add_new_worker(1)

    socketio_client.on("feedback-avatar-updated", mock_avatar_updated_listener)

    worker = game_api.worker_manager.player_id_to_worker[1]
    worker.has_code_updated = False

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await game_api.send_updates_to_all()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)
    await socketio_client.disconnect()

    mock_avatar_updated_listener.assert_not_called()


async def test_remove_session_id_on_disconnect(game_api, client, socketio_server, loop):
    socketio_client = socketio.AsyncClient(reconnection=False)

    game_api.worker_manager.add_new_worker(1)

    await socketio_client.connect(
        f"http://{client.server.host}:{client.server.port}?avatar_id=1&EIO=3&transport=polling&t=MJhoMgb"
    )

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)

    assert len(socketio_server.eio.sockets) == 1

    await socketio_client.disconnect()

    await socketio_server.sleep(TIME_TO_PROCESS_SOME_EVENT_LOOP)

    assert len(socketio_server.eio.sockets) == 0



from simulation.interactables.conditions import avatar_on_cell
from simulation.interactables.effects import ScoreEffect
from simulation.interactables.interactable import _Interactable


class ScoreLocation(_Interactable):
    def __init__(self, cell):
        super(ScoreLocation, self).__init__(cell)

        self.conditions.append(avatar_on_cell)
        self.effects.append(ScoreEffect)

    def get_targets(self):
        return [self.cell.avatar]

    def __repr__(self):
        return f"ScoreLocation(Location={self.cell.location})"

    def serialize(self):
        return {
            "type": "score",
            "location": {"x": self.cell.location.x, "y": self.cell.location.y},
        }



from __future__ import absolute_import
from django.contrib import admin
from .models import Avatar, Game, User
import uuid

from .views import _create_avatar_for_user

NUMBER_OF_AVATARS_TO_ADD = 10


def create_avatar_for_game(game):
    user = User.objects.create_user(uuid.uuid4())
    user.save()
    _create_avatar_for_user(user, game.id, "random_avatar")
    return user


def add_test_avatars_to_games(game_data_admin, request, queryset):
    for game in queryset:
        users = [create_avatar_for_game(game) for _ in range(NUMBER_OF_AVATARS_TO_ADD)]
        game.can_play.add(*users)
        game.save()


add_test_avatars_to_games.short_description = "Add {} test avatars to selected games".format(
    NUMBER_OF_AVATARS_TO_ADD
)


class GameDataAdmin(admin.ModelAdmin):
    list_display = ["id", "name"]

    readonly_fields = ["players"]

    actions = [add_test_avatars_to_games]

    def players(self, obj):
        return "\n".join([u.first_name for u in obj.can_play.all()])


class AvatarDataAdmin(admin.ModelAdmin):
    list_display = ["id", "owner_name", "game_name"]

    def owner_name(self, obj):
        return obj.owner

    def game_name(self, obj):
        return obj.game


admin.site.register(Game, GameDataAdmin)
admin.site.register(Avatar, AvatarDataAdmin)



from django import template

register = template.Library()


@register.inclusion_tag("players/dropdown.html", takes_context=True)
def game_dropdown_list(context, base_url):
    return get_user_playable_games(context, base_url)


def get_user_playable_games(context, base_url):
    user = context.request.user
    return {
        "base_url": base_url,
        "open_play_games": user.playable_games.all()
        if user.is_authenticated()
        else None,
    }



import itertools
import json
import logging
import os

import docker
from docker.models.containers import Container

from .worker import Worker

LOGGER = logging.getLogger(__name__)

port_counter = itertools.count(1989 + int(os.environ.get("GAME_ID", 0)) * 10000)


class LocalWorker(Worker):
    """Relies on them already being created already."""

    host = os.environ.get("LOCALHOST_IP", "127.0.0.1")
    worker_directory = os.path.join(
        os.path.dirname(__file__), "../../../aimmo-game-worker/"
    )

    def __init__(self, *args, **kwargs):
        self.game_id = os.environ.get("GAME_ID", 0)
        self.client = docker.from_env()
        self.container: Container = None
        super(LocalWorker, self).__init__(*args, **kwargs)

    @staticmethod
    def _init_port_counter():
        global port_counter
        port_counter = itertools.count(1989 + int(os.environ["GAME_ID"]) * 10000)

    def _create_worker(self):
        global port_counter
        port = next(port_counter)
        template_string = os.environ.get("CONTAINER_TEMPLATE")
        if template_string:
            template = json.loads(template_string)
        else:
            template = {"environment": {}}
        data_url = "http://{}:{}/player/{}".format(
            self.host, self.game_port, self.player_id
        )
        template["environment"]["DATA_URL"] = data_url
        template["environment"]["PORT"] = port
        self.container = self.client.containers.run(
            name="aimmo-{}-worker-{}".format(self.game_id, self.player_id),
            image="ocadotechnology/aimmo-game-worker:test",
            ports={f"{port}/tcp": port},
            **template,
        )
        worker_url = "http://%s:%d" % (self.host, port)
        LOGGER.info(
            "Worker started for %s, listening at %s", self.player_id, worker_url
        )
        return worker_url

    def remove_worker(self):
        self.container.kill()



import os

dirname = os.path.dirname
abspath = os.path.abspath

db_name = os.path.join(abspath(dirname(dirname(abspath(__file__)))), "db.sqlite3")

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": db_name}}

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "aimmo",
    "aimmo_runner",
]

PIPELINE_ENABLED = False
ROOT_URLCONF = "django_autoconfig.autourlconf"
STATIC_ROOT = ".tests_static/"
SECRET_KEY = "test_key"
GAME_SERVER_URL_FUNCTION = lambda game_id: ("base %s" % game_id, "path %s" % game_id)
GAME_SERVER_PORT_FUNCTION = lambda game_id: 8001
GAME_SERVER_SSL_FLAG = False
from django_autoconfig.autoconfig import configure_settings  # noqa: E402

configure_settings(globals())



from __future__ import absolute_import

import math
from string import ascii_uppercase
from unittest import TestCase

from simulation.game_logic import SpawnLocationFinder
from simulation.interactables.interactable import _Interactable
from simulation.interactables.pickups import ALL_PICKUPS, HealthPickup
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location
from simulation.world_map import WorldMap, WorldMapStaticSpawnDecorator

from .dummy_avatar import DummyAvatar
from .maps import AvatarMap, MockCell, MockPickup


def int_ceil(num):
    return int(math.ceil(num))


def int_floor(num):
    return int(math.floor(num))


class serializer(object):
    def __init__(self, content):
        self.content = content

    def serialize(self):
        return self.content


class TestWorldMap(TestCase):
    def setUp(self):
        self.settings = {
            "TARGET_NUM_CELLS_PER_AVATAR": 0,
            "TARGET_NUM_PICKUPS_PER_AVATAR": 0,
            "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": 0,
            "SCORE_DESPAWN_CHANCE": 0,
            "PICKUP_SPAWN_CHANCE": 0,
        }

    def _generate_grid(self, columns=2, rows=2):
        alphabet = iter(ascii_uppercase)
        grid = {
            Location(x, y): MockCell(Location(x, y), name=next(alphabet))
            for x in range(columns)
            for y in range(rows)
        }
        return grid

    def _grid_from_list(self, in_list):
        out = {}
        for x, column in enumerate(in_list):
            for y, cell in enumerate(column):
                out[Location(x, y)] = cell
        return out

    def assertGridSize(self, world_map, expected_columns, expected_rows=None):
        if expected_rows is None:
            expected_rows = expected_columns
        self.assertEqual(world_map.num_rows, expected_rows)
        self.assertEqual(world_map.num_cols, expected_columns)
        self.assertEqual(world_map.num_cells, expected_rows * expected_columns)
        self.assertEqual(
            len(list(world_map.all_cells())), expected_rows * expected_columns
        )

    def test_get_all_cells(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        cell_list = list(world_map.all_cells())

        self.assertEqual(len(cell_list), 4)
        self.assertTrue(isinstance(cell_list[0], MockCell))

    def test_get_all_interactables(self):
        interactable_cell = MockCell()
        interactable_cell.interactable = ScoreLocation(interactable_cell)
        grid = self._grid_from_list(
            [[interactable_cell, MockCell()], [MockCell(), MockCell()]]
        )
        world_map = WorldMap(grid, self.settings)
        interactable_list = list(world_map.interactable_cells())

        self.assertEqual(len(interactable_list), 1)
        self.assertTrue(isinstance(interactable_list[0].interactable, _Interactable))

    def test_get_all_score_locations(self):
        score_cell = MockCell()
        score_cell.interactable = ScoreLocation(score_cell)
        grid = self._grid_from_list(
            [[score_cell, MockCell()], [MockCell(), MockCell()]]
        )
        world_map = WorldMap(grid, self.settings)
        score_list = list(world_map.score_cells())

        self.assertEqual(len(score_list), 1)
        self.assertTrue(isinstance(score_list[0].interactable, ScoreLocation))

    def test_get_all_pickup_locations(self):
        pickup_cell = MockCell()
        pickup_cell.interactable = HealthPickup(pickup_cell)
        grid = self._grid_from_list(
            [[pickup_cell, MockCell()], [MockCell(), MockCell()]]
        )
        world_map = WorldMap(grid, self.settings)
        pickup_list = list(world_map.pickup_cells())

        self.assertEqual(len(pickup_list), 1)
        self.assertTrue(isinstance(pickup_list[0].interactable, ALL_PICKUPS))

    def test_grid_size(self):
        world_map = WorldMap(self._generate_grid(1, 3), self.settings)
        self.assertGridSize(world_map, 1, 3)

    def test_generated_map(self):
        world_map = WorldMap.generate_empty_map(2, 5, {})
        self.assertGridSize(world_map, 5, 2)

    def test_all_cells(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        cell_names = [c.name for c in world_map.all_cells()]
        self.assertIn("A", cell_names)
        self.assertIn("B", cell_names)
        self.assertIn("C", cell_names)
        self.assertIn("D", cell_names)
        self.assertEqual(len(cell_names), 4)

    def test_score_cells(self):
        score_cell1 = MockCell()
        score_cell1.interactable = ScoreLocation(score_cell1)
        score_cell2 = MockCell()
        score_cell2.interactable = ScoreLocation(score_cell2)
        no_score_cell = MockCell()
        grid = self._grid_from_list(
            [[score_cell1, no_score_cell], [no_score_cell, score_cell2]]
        )
        world_map = WorldMap(grid, self.settings)
        cells = list(world_map.score_cells())
        self.assertIn(score_cell1, cells)
        self.assertIn(score_cell2, cells)
        self.assertEqual(len(cells), 2, "Non-scoring cells present")

    def test_potential_spawns(self):
        spawnable1 = MockCell()
        spawnable2 = MockCell()
        score_cell = MockCell()
        score_cell.interactable = ScoreLocation(score_cell)
        unhabitable = MockCell(habitable=False)
        filled = MockCell(avatar="avatar")
        grid = self._grid_from_list(
            [[spawnable1, score_cell, unhabitable], [unhabitable, spawnable2, filled]]
        )
        world_map = WorldMap(grid, self.settings)
        spawn_location_finder = SpawnLocationFinder(world_map)
        cells = list(spawn_location_finder.potential_spawn_locations())
        self.assertIn(spawnable1, cells)
        self.assertIn(spawnable2, cells)
        self.assertNotIn(score_cell, cells, "Score cells should not be spawns")
        self.assertNotIn(unhabitable, cells, "Unhabitable cells should not be spawns")
        self.assertNotIn(filled, cells, "Cells with avatars should not be spawns")
        self.assertEqual(len(cells), 2)

    def test_interactable_cells(self):
        pickup_cell1 = MockCell(interactable=MockPickup())
        pickup_cell2 = MockCell(interactable=MockPickup())
        no_pickup_cell = MockCell()
        grid = self._grid_from_list(
            [[pickup_cell1, no_pickup_cell], [no_pickup_cell, pickup_cell2]]
        )
        world_map = WorldMap(grid, self.settings)
        cells = list(world_map.interactable_cells())
        self.assertIn(pickup_cell1, cells)
        self.assertIn(pickup_cell2, cells)
        self.assertEqual(len(cells), 2, "Non-pickup cells present")

    def test_location_on_map(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for x in (0, 1):
            for y in (0, 1):
                self.assertTrue(world_map.is_on_map(Location(x, y)))

        self.assertFalse(world_map.is_on_map(Location(2, 2)))
        self.assertFalse(world_map.is_on_map(Location(-1, 1)))

    def test_x_off_map(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for y in (0, 1):
            self.assertFalse(world_map.is_on_map(Location(-1, y)))
            self.assertFalse(world_map.is_on_map(Location(2, y)))

    def test_y_off_map(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for x in (0, 1):
            self.assertFalse(world_map.is_on_map(Location(x, -1)))
            self.assertFalse(world_map.is_on_map(Location(x, 2)))

    def test_get_valid_cell(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for x in (0, 1):
            for y in (0, 1):
                location = Location(x, y)
                self.assertEqual(world_map.get_cell(location).location, location)

    def test_get_x_off_map(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for y in (0, 1):
            with self.assertRaises(ValueError):
                world_map.get_cell(Location(-1, y))
            with self.assertRaises(ValueError):
                world_map.get_cell(Location(2, y))

    def test_get_y_off_map(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for x in (0, 1):
            with self.assertRaises(ValueError):
                world_map.get_cell(Location(x, -1))
            with self.assertRaises(ValueError):
                world_map.get_cell(Location(x, 2))

    def test_random_spawn_location_successful(self):
        cell = MockCell()
        world_map = WorldMap({Location(0, 0): cell}, self.settings)
        self.assertEqual(world_map.get_random_spawn_location(), cell.location)

    def test_random_spawn_location_with_no_candidates(self):
        grid = self._generate_grid(1, 1)
        world_map = WorldMap(grid, self.settings)
        grid[Location(0, 0)].avatar = True
        with self.assertRaises(IndexError):
            world_map.get_random_spawn_location()

    def test_can_move_to(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        target = Location(1, 1)
        self.assertTrue(world_map.can_move_to(target))

    def test_cannot_move_to_cell_off_grid(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        target = Location(4, 1)
        self.assertFalse(world_map.can_move_to(target))

    def test_cannot_move_to_uninhabitable_cell(self):
        target = Location(0, 0)
        cell = MockCell(target, habitable=False)
        world_map = WorldMap({target: cell}, self.settings)
        self.assertFalse(world_map.can_move_to(target))

    def test_cannot_move_to_habited_cell(self):
        target = Location(0, 0)
        cell = MockCell(target, avatar=DummyAvatar(target, 0))
        world_map = WorldMap({target: cell}, self.settings)
        target = Location(0, 0)
        self.assertFalse(world_map.can_move_to(target))

    def test_empty_grid(self):
        world_map = WorldMap({}, self.settings)
        self.assertFalse(world_map.is_on_map(Location(0, 0)))

    def test_iter(self):
        grid = [
            [
                MockCell(Location(-1, -1), name="A"),
                MockCell(Location(-1, 0), name="B"),
                MockCell(Location(-1, 1), name="C"),
            ],
            [
                MockCell(Location(0, -1), name="D"),
                MockCell(Location(0, 0), name="E"),
                MockCell(Location(0, 1), name="F"),
            ],
            [
                MockCell(Location(1, -1), name="E"),
                MockCell(Location(1, 0), name="G"),
                MockCell(Location(1, 1), name="H"),
            ],
        ]
        world_map = WorldMap(self._grid_from_list(grid), self.settings)
        self.assertEqual([list(column) for column in world_map], grid)

    def test_attackable_avatar_returns_none(self):
        world_map = WorldMap(self._generate_grid(), self.settings)
        for x in (0, 1):
            for y in (0, 1):
                self.assertIsNone(world_map.attackable_avatar(Location(x, y)))

    def test_attackable_avatar_returns_avatar(self):
        avatar = DummyAvatar()
        world_map = AvatarMap(avatar)

        self.assertEqual(world_map.attackable_avatar(Location(0, 0)), avatar)


class TestWorldMapWithOriginCentre(TestWorldMap):
    def _generate_grid(self, columns=2, rows=2):
        alphabet = iter(ascii_uppercase)
        grid = {
            Location(x, y): MockCell(Location(x, y), name=next(alphabet))
            for x in range(-int_ceil(columns / 2.0) + 1, int_floor(columns / 2.0) + 1)
            for y in range(-int_ceil(rows / 2.0) + 1, int_floor(rows / 2.0) + 1)
        }

        return grid

    def _grid_from_list(self, in_list):
        out = {}
        min_x = -int_ceil(len(in_list) / 2.0) + 1
        min_y = -int_ceil(len(in_list[0]) / 2.0) + 1
        for i, column in enumerate(in_list):
            x = i + min_x
            for j, cell in enumerate(column):
                y = j + min_y
                out[Location(x, y)] = cell
        return out

    def test_retrieve_negative(self):
        world_map = WorldMap(self._generate_grid(3, 3), self.settings)
        self.assertTrue(world_map.is_on_map(Location(-1, -1)))


class TestStaticSpawnDecorator(TestCase):
    def test_spawn_is_static(self):
        decorated_map = WorldMapStaticSpawnDecorator(WorldMap({}, {}), Location(3, 7))
        for _ in range(5):
            self.assertEqual(decorated_map.get_random_spawn_location(), Location(3, 7))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("aimmo", "0004_auto_20160808_1511"),
    ]

    operations = [
        migrations.CreateModel(
            name="LevelAttempt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("level_number", models.IntegerField()),
            ],
        ),
        migrations.AddField(
            model_name="game",
            name="completed",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="game",
            name="generator",
            field=models.CharField(
                default=b"Main",
                max_length=20,
                choices=[(b"Main", b"Open World"), (b"Level1", b"Level 1")],
            ),
        ),
        migrations.AddField(
            model_name="game",
            name="main_user",
            field=models.ForeignKey(
                related_name="games_for_user",
                blank=True,
                to=settings.AUTH_USER_MODEL,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="game",
            name="owner",
            field=models.ForeignKey(
                related_name="owned_games",
                blank=True,
                to=settings.AUTH_USER_MODEL,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="levelattempt",
            name="game",
            field=models.OneToOneField(to="aimmo.Game"),
        ),
        migrations.AddField(
            model_name="levelattempt",
            name="user",
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterUniqueTogether(
            name="levelattempt", unique_together=set([("level_number", "user")])
        ),
    ]



#!/usr/bin/env python

import asyncio
import json
import logging
import os
import sys
from urllib.parse import parse_qs

import aiohttp_cors
import socketio
from activity_monitor import ActivityMonitor
from aiohttp import web
from aiohttp_wsgi import WSGIHandler
from authentication import initialize_game_token
from prometheus_client import make_wsgi_app
from simulation import map_generator
from simulation.django_communicator import DjangoCommunicator
from simulation.game_runner import GameRunner

LOGGER = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

django_api_url = os.environ.get(
    "GAME_API_URL", "http://localhost:8000/kurono/api/games/"
)

communicator = DjangoCommunicator(
    django_api_url=django_api_url, completion_url=django_api_url + "complete/"
)
activity_monitor = ActivityMonitor(communicator)


def setup_application(should_clean_token=True):
    async def clean_token(app):
        LOGGER.info("Cleaning token!")
        await communicator.patch_token(data={"token": ""})

    application = web.Application()

    if should_clean_token:
        application.on_shutdown.append(clean_token)

    application.on_shutdown.append(communicator.close_session)

    return application


def setup_prometheus():
    wsgi_handler = WSGIHandler(make_wsgi_app())
    app.add_routes([web.get("/{path_info:metrics}", wsgi_handler)])


def setup_socketIO_server(
    application, client_manager_class=socketio.AsyncManager, async_handlers=True
):
    socket_server = socketio.AsyncServer(
        client_manager=client_manager_class(), async_handlers=async_handlers
    )

    socket_server.attach(
        application, socketio_path=os.environ.get("SOCKETIO_RESOURCE", "socket.io")
    )

    return socket_server


app = setup_application()
cors = aiohttp_cors.setup(app)

socketio_server = setup_socketIO_server(app)


class GameAPI(object):

    routes = web.RouteTableDef()

    def __init__(
        self, game_state, worker_manager, application=app, server=socketio_server
    ):
        self.app = application
        self.socketio_server = server
        self.register_endpoints()
        self.worker_manager = worker_manager
        self.game_state = game_state

    async def async_map(self, func, iterable_args):
        futures = [func(arg) for arg in iterable_args]
        await asyncio.gather(*futures)

    def register_endpoints(self):
        self.register_player_data_view()
        self.register_world_update_on_connect()
        self.register_remove_session_id_from_mappings()
        self.register_healthcheck()
        self.app.add_routes(self.routes)

    def open_connections_number(self):
        try:
            return len(
                [x for x in self.socketio_server.manager.get_participants("/", None)]
            )
        except KeyError:
            return 0

    def update_active_users(self):
        activity_monitor.active_users = self.open_connections_number()

    def register_healthcheck(self):
        @self.routes.get("/game-{game_id}")
        async def healthcheck(request):
            return web.Response(text="HEALTHY")

        return healthcheck

    def register_player_data_view(self):
        @self.routes.get("/player/{player_id}")
        async def player_data(request: web.Request):
            player_id = int(request.match_info["player_id"])
            return web.json_response(
                {
                    "code": self.worker_manager.get_code(player_id),
                    "options": {},
                    "state": None,
                }
            )

        return player_data

    def register_world_update_on_connect(self):
        @self.socketio_server.on("connect")
        async def world_update_on_connect(sid, environ):
            LOGGER.info(f"Socket connected for session id: {sid}")
            query = environ["QUERY_STRING"]
            avatar_id = self._find_avatar_id_from_query(sid, query)
            await self.socketio_server.save_session(sid, {"id": avatar_id})

        return world_update_on_connect

    def register_remove_session_id_from_mappings(self):
        @self.socketio_server.on("disconnect")
        async def remove_session_id_from_mappings(sid):
            LOGGER.info("Socket disconnected for session id: {}. ".format(sid))

        return remove_session_id_from_mappings

    async def send_updates(self, sid):
        try:
            await self._send_have_avatars_code_updated(sid)
            await self._send_game_state(sid)
            await self._send_logs(sid)
        except KeyError:
            LOGGER.error(
                f"Failed to send updates. No worker for player in session {sid}"
            )

    async def send_updates_to_all(self):
        try:
            socket_ids = self.socketio_server.manager.get_participants("/", None)
            await self.async_map(self.send_updates, socket_ids)
        except KeyError:
            LOGGER.error("No open socket connections")
        self.update_active_users()

    def _find_avatar_id_from_query(self, session_id, query_string):
        """
        :param session_id: Int with the session id
        :param query_string: String from the environment settings,
        usually located as the key 'QUERY_STRING'.
        """
        parsed_qs = parse_qs(query_string)

        try:
            avatar_id = int(parsed_qs["avatar_id"][0])
            return avatar_id
        except ValueError:
            LOGGER.error("Avatar ID could not be casted into an integer")
        except KeyError:
            LOGGER.error("No avatar ID found. User may not be authorised")
            LOGGER.error("query_string: " + query_string)

    async def _send_logs(self, sid):
        def should_send_logs(logs):
            return bool(logs)

        session_data = await self.socketio_server.get_session(sid)
        worker = self.worker_manager.player_id_to_worker[session_data["id"]]
        avatar_logs = worker.log

        if should_send_logs(avatar_logs):
            await self.socketio_server.emit(
                "log",
                {"message": avatar_logs, "turn_count": self.game_state.turn_count},
                room=sid,
            )

    async def _send_game_state(self, sid):
        serialized_game_state = self.game_state.serialize()
        session_data = await self.socketio_server.get_session(sid)
        worker = self.worker_manager.player_id_to_worker[session_data["id"]]
        if worker.ready:
            await self.socketio_server.emit(
                "game-state", serialized_game_state, room=sid
            )

    async def _send_have_avatars_code_updated(self, sid):
        session_data = await self.socketio_server.get_session(sid)
        worker = self.worker_manager.player_id_to_worker[session_data["id"]]
        if worker.has_code_updated:
            await self.socketio_server.emit("feedback-avatar-updated", {}, room=sid)


def create_runner(port):
    settings = json.loads(os.environ["settings"])
    generator = getattr(map_generator, settings["GENERATOR"])(settings)
    return GameRunner(
        game_state_generator=generator.get_game_state,
        communicator=communicator,
        port=port,
    )


def run_game(port):
    game_runner = create_runner(port)

    game_api = GameAPI(
        game_state=game_runner.game_state, worker_manager=game_runner.worker_manager
    )
    game_runner.set_end_turn_callback(game_api.send_updates_to_all)
    asyncio.ensure_future(game_runner.run())


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    host = sys.argv[1]

    if os.environ["WORKER"] == "local":
        port = int(os.environ["EXTERNAL_PORT"])
    else:
        port = int(sys.argv[2])

    asyncio.ensure_future(initialize_game_token(communicator))
    run_game(port)

    setup_prometheus()

    logging.getLogger("socketio").setLevel(logging.ERROR)
    logging.getLogger("engineio").setLevel(logging.ERROR)
    LOGGER.info("starting the server")
    web.run_app(app, host=host, port=port)



import json

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import Client, TestCase
from rest_framework import status

from aimmo import app_settings, models
from aimmo.serializers import GameSerializer

app_settings.GAME_SERVER_URL_FUNCTION = lambda game_id: (
    "base %s" % game_id,
    "path %s" % game_id,
)
app_settings.GAME_SERVER_PORT_FUNCTION = lambda game_id: 0
app_settings.GAME_SERVER_SSL_FLAG = True
DEFAULT_CODE = """def next_turn(world_state, avatar_state):
    return MoveAction(direction.NORTH)
"""


class TestViews(TestCase):
    CODE = "class Avatar: pass"

    EXPECTED_GAMES = {
        "main_avatar": 1,
        "users": [
            {"id": 1, "code": CODE},
            {"id": 2, "code": "test2"},
            {"id": 3, "code": "test3"},
        ],
    }

    EXPECTED_GAME_DETAIL = {
        "name": "test",
        "status": "r",
        "settings": '{"GENERATOR": "Main", "OBSTACLE_RATIO": 0.1, "PICKUP_SPAWN_CHANCE": 0.1, "SCORE_DESPAWN_CHANCE": 0.05, "START_HEIGHT": 31, "START_WIDTH": 31, "TARGET_NUM_CELLS_PER_AVATAR": 16.0, "TARGET_NUM_PICKUPS_PER_AVATAR": 1.2, "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": 0.5}',
    }

    EXPECTED_GAME_LIST = {"1": EXPECTED_GAME_DETAIL, "2": EXPECTED_GAME_DETAIL}

    @classmethod
    def setUpTestData(cls):
        cls.user = User.objects.create_user("test", "test@example.com", "password")
        cls.user.is_staff = True
        cls.user.save()
        cls.game = models.Game(id=1, name="test", public=True)
        cls.game.save()

    def setUp(self):
        self.game.refresh_from_db()

    def login(self):
        c = Client()
        c.login(username="test", password="password")
        return c

    def _go_to_page(self, link, kwarg_name, game_id):
        c = Client()
        response = c.get(reverse(link, kwargs={kwarg_name: game_id}))
        return response

    def _make_game_private(self):
        self.game.public = False
        self.game.save()

    def test_add_new_code(self):
        c = self.login()
        response = c.post(reverse("kurono/code", kwargs={"id": 1}), {"code": self.CODE})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            models.Avatar.objects.get(owner__username="test").code, self.CODE
        )

    def test_update_code(self):
        c = self.login()
        models.Avatar(owner=self.user, code="test", game=self.game).save()
        response = c.post(reverse("kurono/code", kwargs={"id": 1}), {"code": self.CODE})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            models.Avatar.objects.get(owner__username="test").code, self.CODE
        )

    def test_code_for_non_existant_game(self):
        c = self.login()
        response = c.post(reverse("kurono/code", kwargs={"id": 2}), {"code": self.CODE})
        self.assertEqual(response.status_code, 404)

    def test_code_for_non_authed_user(self):
        self.game.public = False
        self.game.save()
        c = self.login()
        response = c.post(reverse("kurono/code", kwargs={"id": 1}), {"code": self.CODE})
        self.assertEqual(response.status_code, 404)

    def test_default_code(self):
        c = self.login()
        response = c.get(reverse("kurono/code", kwargs={"id": 1}))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(DEFAULT_CODE, json.loads(response.content)["code"])

    def test_retrieve_code(self):
        models.Avatar(owner=self.user, code=self.CODE, game=self.game).save()
        c = self.login()
        response = c.get(reverse("kurono/code", kwargs={"id": 1}))
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content, {"code": self.CODE})

    def _associate_game_as_level_num(self, level_num=1, user=None, game=None):
        if game is None:
            game = self.game
        if user is None:
            user = self.user
        models.LevelAttempt(user=user, game=game, level_number=level_num).save()
        models.Game(name="Wrong").save()

    def test_play(self):
        c = self.login()
        response = c.get(reverse("kurono/play", kwargs={"id": 1}))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context["current_user_player_key"], self.user.pk)
        self.assertEqual(response.context["game_url_base"], "base 1")
        self.assertEqual(response.context["game_url_path"], "path 1")

    def test_play_for_non_existent_game(self):
        c = self.login()
        response = c.get(reverse("kurono/play", kwargs={"id": 2}))
        self.assertEqual(response.status_code, 404)

    def _run_test_for_unauthorised_user(self, link, kwarg_name, status_code):
        self._make_game_private()
        c = self.login()
        response = c.get(reverse(link, kwargs={kwarg_name: 1}))
        self.assertEqual(response.status_code, status_code)

    def test_play_for_unauthorised_user(self):
        self._run_test_for_unauthorised_user("kurono/play", "id", 404)

    def test_play_inactive_level(self):
        c = self.login()
        self.game.completed = True
        if self.game.is_active:
            self.skipTest("Completed game is active")
        self.game.static_data = '{"test": 1}'
        self.game.save()
        response = c.get(reverse("kurono/play", kwargs={"id": 1}))
        self.assertEqual(response.status_code, 200)
        self.assertFalse(response.context["active"])
        self.assertEqual(response.context["static_data"], '{"test": 1}')

    def test_games_api(self):
        self.game.main_user = self.user
        self.game.save()
        user2 = User.objects.create_user(username="2", password="password")
        user3 = User.objects.create_user(username="3", password="password")
        models.Avatar(owner=self.user, code=self.CODE, pk=1, game=self.game).save()
        models.Avatar(owner=user2, code="test2", pk=2, game=self.game).save()
        models.Avatar(owner=user3, code="test3", pk=3, game=self.game).save()
        c = Client()
        response = c.get(reverse("kurono/game_user_details", kwargs={"id": 1}))
        self.assertJSONEqual(response.content, self.EXPECTED_GAMES)

    def test_games_api_for_non_existent_game(self):
        response = self._go_to_page("kurono/game_user_details", "id", 5)
        self.assertEqual(response.status_code, 404)

    def _run_mark_complete_test(self, request_method, game_id, success_expected):
        c = Client()
        if request_method == "POST":
            response = c.post(reverse("kurono/complete_game", kwargs={"id": game_id}))
        else:
            response = c.get(reverse("kurono/complete_game", kwargs={"id": game_id}))
        self.assertEqual(response.status_code == 200, success_expected)
        self.assertEqual(models.Game.objects.get(id=1).completed, success_expected)

    def test_mark_complete(self):
        self._run_mark_complete_test("POST", 1, True)

    def test_mark_complete_for_non_existent_game(self):
        self._run_mark_complete_test("POST", 3, False)

    def test_mark_complete_requires_POST(self):
        self._run_mark_complete_test("GET", 1, False)

    def test_mark_complete_has_no_csrf_check(self):
        c = Client(enforce_csrf_checks=True)
        response = c.post(reverse("kurono/complete_game", kwargs={"id": 1}))
        self.assertEqual(response.status_code, 200)
        self.assertTrue(models.Game.objects.get(id=1).completed)

    def test_mark_complete_with_data(self):
        c = Client()
        c.post(
            reverse("kurono/complete_game", kwargs={"id": 1}),
            "static",
            content_type="application/json",
        )
        self.assertEqual(models.Game.objects.get(id=1).static_data, "static")

    def test_stop_game(self):
        game = models.Game.objects.get(id=1)
        game.auth_token = "tokenso lorenzo"
        game.save()
        c = Client()

        response = c.patch(
            reverse("game-detail", kwargs={"pk": 1}),
            json.dumps({"status": models.Game.STOPPED}),
            content_type="application/json",
            HTTP_GAME_TOKEN=game.auth_token,
        )
        game = models.Game.objects.get(id=1)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(game.status, models.Game.STOPPED)

    def test_stop_game_no_token(self):
        game = models.Game.objects.get(id=1)
        game.auth_token = "tokenso lorenzo"
        game.save()
        c = Client()

        response = c.patch(
            reverse("game-detail", kwargs={"pk": 1}),
            json.dumps({"status": models.Game.STOPPED}),
            content_type="application/json",
        )
        game = models.Game.objects.get(id=1)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(game.status, models.Game.RUNNING)

    def test_current_avatar_api_for_non_existent_game(self):
        response = self._go_to_page("kurono/current_avatar_in_game", "game_id", 1)
        self.assertEqual(response.status_code, 404)

    def test_current_avatar_api_for_unauthorised_games(self):
        self._run_test_for_unauthorised_user(
            "kurono/current_avatar_in_game", "game_id", 401
        )

    def test_current_avatar_api_for_two_users(self):
        # Set up the first avatar
        first_user = self.user
        models.Avatar(owner=first_user, code=self.CODE, game=self.game).save()
        client_one = self.login()

        # Set up the second avatar
        second_user = User.objects.create_user(
            "test2", "test2@example.com", "password2"
        )
        second_user.save()
        models.Avatar(owner=second_user, code=self.CODE, game=self.game).save()
        client_two = Client()
        client_two.login(username="test2", password="password2")

        self.game.public = True
        self.game.can_play = [first_user, second_user]
        self.game.save()

        first_response = client_one.get(
            reverse("kurono/current_avatar_in_game", kwargs={"game_id": 1})
        )
        second_response = client_two.get(
            reverse("kurono/current_avatar_in_game", kwargs={"game_id": 1})
        )

        # Status code starts with 2, success response can be different than 200.
        self.assertEqual(str(first_response.status_code)[0], "2")
        self.assertEqual(str(second_response.status_code)[0], "2")

        # JSON is returned as string so needs to be evaluated.
        first_id = json.loads(first_response.content)["current_avatar_id"]
        second_id = json.loads(second_response.content)["current_avatar_id"]

        self.assertEqual(first_id, 1)
        self.assertEqual(second_id, 2)

    def test_current_avatar_api_call_returns_404_for_logged_out_user(self):
        user = self.user
        models.Avatar(owner=user, code=self.CODE, game=self.game).save()
        client_one = Client()

        self.game.public = True
        self.game.can_play = [user]
        self.game.save()

        first_response = client_one.get(
            reverse("kurono/current_avatar_in_game", kwargs={"game_id": 1})
        )

        self.assertEqual(first_response.status_code, 404)

    def test_id_of_current_avatar_same_as_games_url(self):
        """
        Ensures that the id's are consistent throughout the project. Check for ID's received
        by the current_avatar URL as well as the games URL api.
        """
        user = self.user
        models.Avatar(owner=user, code=self.CODE, game=self.game).save()
        client = self.login()

        self.game.public = True
        self.game.can_play = [user]
        self.game.save()

        current_avatar_api_response = client.get(
            reverse("kurono/current_avatar_in_game", kwargs={"game_id": 1})
        )
        games_api_response = client.get(
            reverse("kurono/game_user_details", kwargs={"id": 1})
        )

        current_avatar_id = json.loads(current_avatar_api_response.content)[
            "current_avatar_id"
        ]
        games_api_users = json.loads(games_api_response.content)["users"]

        self.assertEqual(current_avatar_id, 1)
        self.assertEqual(len(games_api_users), 1)
        self.assertEqual(games_api_users[0]["id"], 1)

    def test_token_view_get_token_multiple_requests(self):
        """
        Ensures we can make a get request for the token, and
        that a request with a valid token is also accepted.
        """
        token = models.Game.objects.get(id=1).auth_token
        client = Client()
        response = client.get(reverse("kurono/game_token", kwargs={"id": 1}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(token, response.json()["token"])

        # Token starts as empty, as long as it is empty, we can make more GET requests
        response = client.get(reverse("kurono/game_token", kwargs={"id": 1}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(token, response.json()["token"])

    def test_get_token_after_token_set(self):
        token = models.Game.objects.get(id=1).auth_token
        client = Client()
        response = client.get(reverse("kurono/game_token", kwargs={"id": 1}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(token, response.json()["token"])

        new_token = "aaaaaaaaaaa"
        response = client.patch(
            reverse("kurono/game_token", kwargs={"id": 1}),
            json.dumps({"token": new_token}),
            content_type="application/json",
        )

        # Token starts as empty, as long as it is empty, we can make more GET requests
        response = client.get(
            reverse("kurono/game_token", kwargs={"id": 1}), HTTP_GAME_TOKEN=new_token
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_patch_token_with_no_token(self):
        """
        Check for 401 when attempting to change game token.
        """
        client = Client()
        token = models.Game.objects.get(id=1).auth_token
        response = client.patch(reverse("kurono/game_token", kwargs={"id": 1}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_patch_token_with_incorrect_token(self):
        """
        Check for 403 when attempting to change game token (incorrect token provided).
        """
        client = Client()
        token = models.Game.objects.get(id=1).auth_token
        response = client.patch(
            reverse("kurono/game_token", kwargs={"id": 1}),
            {},
            content_type="application/json",
            HTTP_GAME_TOKEN="INCORRECT TOKEN",
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_patch_token_with_correct_token(self):
        """
        Check for 200 and successful token change when updating the token (correct token provided).
        """
        client = Client()
        token = models.Game.objects.get(id=1).auth_token
        new_token = token[::-1]
        response = client.patch(
            reverse("kurono/game_token", kwargs={"id": 1}),
            json.dumps({"token": new_token}),
            content_type="application/json",
            HTTP_GAME_TOKEN=token,
        )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(models.Game.objects.get(id=1).auth_token, new_token)

    def test_delete_game(self):
        """
        Check for 204 when deleting a game
        """
        client = self.login()

        game2 = models.Game(id=2, name="test", public=True)
        game2.save()

        response = client.delete(reverse("game-detail", kwargs={"pk": self.game.id}))
        self.assertEquals(response.status_code, 204)
        self.assertEquals(len(models.Game.objects.all()), 1)

    def test_delete_non_existent_game(self):
        c = self.login()
        response = c.delete(reverse("game-detail", kwargs={"pk": 2}))
        self.assertEqual(response.status_code, 404)

    def test_delete_for_unauthorized_user(self):
        """
        Check for 403 when attempting to delete a game without being authorized
        """
        self._make_game_private()
        c = self.login()
        response = c.delete(reverse("game-detail", kwargs={"pk": self.game.id}))
        self.assertEqual(response.status_code, 403)

    def test_game_serializer_settings(self):
        """
        Check that the serializer gets the correct settings data from the game
        """
        client = self.login()

        serializer = GameSerializer(self.game)

        self.assertEquals(
            json.dumps(self.game.settings_as_dict(), sort_keys=True),
            serializer.data["settings"],
        )

    def test_list_all_games(self):
        self.game.main_user = self.user
        self.game.save()

        game2 = models.Game(id=2, name="test", public=True)
        game2.save()

        c = Client()
        response = c.get(reverse("game-list"))

        self.assertJSONEqual(response.content, self.EXPECTED_GAME_LIST)

    def test_view_one_game(self):
        client = self.login()

        response = client.get(reverse("game-detail", kwargs={"pk": self.game.id}))
        self.assertEquals(response.status_code, 200)

    def test_add_game(self):
        client = self.login()

        response = client.post(reverse("kurono/new_game"), data={"name": "test"})
        self.assertEquals(response.status_code, 302)

    def test_adding_a_game_creates_an_avatar(self):
        client = self.login()
        response = client.post(reverse("kurono/new_game"), data={"name": "test"})
        game = models.Game.objects.get(pk=2)
        avatar = game.avatar_set.get(owner=client.session["_auth_user_id"])
        self.assertIsNotNone(avatar)



from functools import wraps

from django.conf import settings
from django.utils.module_loading import import_string
from django.contrib.auth.models import User
from rest_framework import permissions

#: URL function for locating the game server, takes one parameter `game`
GAME_SERVER_URL_FUNCTION = getattr(settings, "AIMMO_GAME_SERVER_URL_FUNCTION", None)
GAME_SERVER_PORT_FUNCTION = getattr(settings, "AIMMO_GAME_SERVER_PORT_FUNCTION", None)
GAME_SERVER_SSL_FLAG = getattr(settings, "AIMMO_GAME_SERVER_SSL_FLAG", False)
CAN_DELETE_GAME_CLASS = getattr(settings, "CAN_DELETE_GAME_CLASS", None)
USERS_FOR_NEW_AIMMO_GAME = getattr(settings, "USERS_FOR_NEW_AIMMO_GAME", None)


class DummyPermission(permissions.BasePermission):
    """
    Used to mock general permissions
    """

    def has_permission(self, request, view):
        return True


def get_users_for_new_game(request):
    """
    Imports and calls a function defined in portal, that decides
    which users should be added to the newly created game.
    :param request:
    :return: List of User objects
    """
    if USERS_FOR_NEW_AIMMO_GAME:
        func = import_string(USERS_FOR_NEW_AIMMO_GAME)
        return func(request)
    return User.objects.all()


def get_can_delete_game_class():
    if CAN_DELETE_GAME_CLASS:
        return import_string(CAN_DELETE_GAME_CLASS)
    return DummyPermission


CanDelete = get_can_delete_game_class()

MAX_LEVEL = 1



import base64
import logging
import os

import kubernetes
import requests
from requests import codes

LOGGER = logging.getLogger(__name__)


def _decode_token_from_secret(secret):
    return base64.b64decode(secret.data["token"]).decode()


async def initialize_game_token(communicator):
    """Get game token and store it somewhere accessible."""
    if os.environ["WORKER"] == "kubernetes":
        kubernetes.config.load_incluster_config()
        api = kubernetes.client.CoreV1Api()
        game_id = os.environ.get("GAME_ID")

        secret = api.read_namespaced_secret(f"game-{game_id}-token", "default")
        os.environ["TOKEN"] = _decode_token_from_secret(secret)
        LOGGER.info("Token set!")

    response = await communicator.patch_token({"token": os.environ["TOKEN"]})
    if response.status != codes["ok"]:
        LOGGER.error(response)



from base64 import urlsafe_b64encode
from os import urandom

from django.contrib.auth.models import User
from django.db import models

from aimmo import app_settings

GAME_GENERATORS = [("Main", "Open World")] + [  # Default
    ("Level%s" % i, "Level %s" % i) for i in range(1, app_settings.MAX_LEVEL + 1)
]


def generate_auth_token():
    return urlsafe_b64encode(urandom(16))


class GameQuerySet(models.QuerySet):
    def for_user(self, user):
        if user.is_authenticated():
            return self.filter(models.Q(public=True) | models.Q(can_play=user))
        else:
            return self.filter(public=True)

    def exclude_inactive(self):
        return self.filter(completed=False)


class Game(models.Model):
    RUNNING = "r"
    STOPPED = "s"
    PAUSED = "p"
    STATUS_CHOICES = ((RUNNING, "running"), (STOPPED, "stopped"), (PAUSED, "paused"))

    name = models.CharField(max_length=100)
    auth_token = models.CharField(max_length=24, blank=True)
    owner = models.ForeignKey(User, blank=True, null=True, related_name="owned_games")
    public = models.BooleanField(default=False)
    can_play = models.ManyToManyField(User, related_name="playable_games")
    completed = models.BooleanField(default=False)
    main_user = models.ForeignKey(
        User, blank=True, null=True, related_name="games_for_user"
    )
    objects = GameQuerySet.as_manager()
    static_data = models.TextField(blank=True, null=True)

    # Game config
    generator = models.CharField(
        max_length=20, choices=GAME_GENERATORS, default=GAME_GENERATORS[0][0]
    )
    target_num_cells_per_avatar = models.FloatField(default=16)
    target_num_score_locations_per_avatar = models.FloatField(default=0.5)
    score_despawn_chance = models.FloatField(default=0.05)
    target_num_pickups_per_avatar = models.FloatField(default=1.2)
    pickup_spawn_chance = models.FloatField(default=0.1)
    obstacle_ratio = models.FloatField(default=0.1)
    start_height = models.IntegerField(default=31)
    start_width = models.IntegerField(default=31)
    status = models.CharField(max_length=1, choices=STATUS_CHOICES, default=RUNNING)

    @property
    def is_active(self):
        return not self.completed

    def __str__(self):
        return self.name

    def can_user_play(self, user):
        return self.public or user in self.can_play.all()

    def settings_as_dict(self):
        return {
            "GENERATOR": self.generator,
            "OBSTACLE_RATIO": self.obstacle_ratio,
            "PICKUP_SPAWN_CHANCE": self.pickup_spawn_chance,
            "SCORE_DESPAWN_CHANCE": self.score_despawn_chance,
            "START_HEIGHT": self.start_height,
            "START_WIDTH": self.start_width,
            "TARGET_NUM_CELLS_PER_AVATAR": self.target_num_cells_per_avatar,
            "TARGET_NUM_PICKUPS_PER_AVATAR": self.target_num_pickups_per_avatar,
            "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": self.target_num_score_locations_per_avatar,
        }

    def save(self, *args, **kwargs):
        super(Game, self).full_clean()
        super(Game, self).save(*args, **kwargs)


class Avatar(models.Model):
    owner = models.ForeignKey(User)
    game = models.ForeignKey(Game)
    code = models.TextField()
    auth_token = models.CharField(max_length=24, default=generate_auth_token)

    class Meta:
        unique_together = ("owner", "game")


class LevelAttempt(models.Model):
    level_number = models.IntegerField()
    user = models.ForeignKey(User)
    game = models.OneToOneField(Game)

    class Meta:
        unique_together = ("level_number", "user")



import os
import logging
import mock
import unittest
from django.test.client import Client
import psutil
from aimmo_runner import runner
from connection_api import (create_session, send_get_request, send_post_request,
                            obtain_csrftoken, delete_old_database, is_server_healthy)
from django.core.urlresolvers import reverse
logging.basicConfig(level=logging.WARNING)


class TestIntegration(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestIntegration, self).__init__(*args, **kwargs)
        self.processes = []

    def tearDown(self):
        """
        Kills the process and its children peacefully.
        """

        for process in self.processes:
            try:
                parent = psutil.Process(process.pid)
            except psutil.NoSuchProcess:
                return

            children = parent.children(recursive=True)

            for child in children:
                child.terminate()

            parent.terminate()

    @mock.patch('docker.from_env')
    def test_superuser_authentication(self, docker_from_env):
        """
        A test that will run on a clean & empty database, create all migrations, new
        browser session and passes a CSRF token with the POST input request.
        
        Server gets killed at the end of the test.
        """
        url_string = 'kurono/login'
        delete_old_database()

        os.chdir(runner.ROOT_DIR_LOCATION)
        self.processes = runner.run(use_minikube=False, server_wait=False, capture_output=True, test_env=True)
        client = Client()
        response = client.get(reverse(url_string))
        self.assertEqual(response.status_code, 200)
        csrf_token = response.context['csrf_token']

        login_info = {
            'username': 'admin',
            'password': 'admin',
            'csrftoken': csrf_token,
        }

        response = client.post(reverse(url_string), login_info)
        self.assertEqual(response.status_code, 302)



import random

from simulation import map_generator
from simulation.simulation_runner import SequentialSimulationRunner
from tests.test_simulation.mock_communicator import MockCommunicator
from tests.test_simulation.dummy_avatar import DummyAvatarManager, MoveEastDummy

SETTINGS = {"START_HEIGHT": 5, "START_WIDTH": 5, "OBSTACLE_RATIO": 0}


class MockWorld(object):
    """
    Creates an object that mocks the whole game and can be used in various testing.
    It holds the map generator, avatar manager, game state and turn manager. Takes settings as a parameter,
    if defaults are unsuitable.

    By default, the first avatar added to the world will be a MoveEastDummy.
    """

    def __init__(
        self,
        settings=SETTINGS,
        dummies_list=None,
        map_generator_class=map_generator.Main,
        simulation_runner_class=SequentialSimulationRunner,
    ):
        random.seed(0)
        if dummies_list is None:
            dummies_list = [MoveEastDummy]

        self.generator = map_generator_class(settings)
        self.avatar_manager = DummyAvatarManager(dummies_list)
        self.game_state = self.generator.get_game_state(self.avatar_manager)
        self.simulation_runner = simulation_runner_class(
            game_state=self.game_state, communicator=MockCommunicator()
        )



import pytest
import asyncio

from simulation.location import Location
from simulation.interactables.pickups import DamageBoostPickup, InvulnerabilityPickup
from simulation.interactables.effects import (
    DAMAGE_BOOST_DEFAULT,
    INVULNERABILITY_RESISTANCE,
)

from .mock_world import MockWorld


class TestEffectExpiry:
    """
    Tests if effects of pickups can expire correctly in multiple scenarios. Note: applies only to pickups that inherit
    TimedEffect!
    """

    def setup_method(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        SETTINGS = {"START_HEIGHT": 5, "START_WIDTH": 15, "OBSTACLE_RATIO": 0}

        self.game = MockWorld(SETTINGS)
        self.game.simulation_runner.add_avatar(1, Location(0, 0))
        self.avatar = self.game.avatar_manager.get_avatar(1)
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))

    async def test_single_damage_boost_pickup_expiry(self, loop):
        """
        Avatar spawns at ORIGIN. DamageBoostPickup is at 1,0. Avatar moves to the pickup and picks it up next turn and
        then we wait for the effect to expire EFFECT_TIME turns later (value defined in the effects class).
        """
        pickup_created = DamageBoostPickup(self.cell)
        self.cell.interactable = pickup_created
        assert self.avatar.attack_strength == 1

        # Avatar moves EAST to (1,0) where pickup is located, then repeats it 5 times.
        for i in range(6):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert isinstance(list(self.avatar.effects)[0], pickup_created.effects[0])
        assert list(self.avatar.effects)[0]._time_remaining == 5
        assert self.avatar.attack_strength == DAMAGE_BOOST_DEFAULT + 1

        # Run 5 more turns and expect the effect to expire.
        for i in range(5):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 0
        assert self.avatar.attack_strength == 1

    async def test_single_invulnerability_pickup_pickup_expiry(self, loop):
        """
        Avatar spawns at ORIGIN. InvulnerabilityPickup is at 1,0. Avatar moves to the pickup and picks it up next turn
        and then we wait for the effect to expire EFFECT_TIME turns later (value defined in the effects class).
        """
        pickup_created = InvulnerabilityPickup(self.cell)
        self.cell.interactable = pickup_created
        assert self.avatar.resistance == 0

        # Avatar moves EAST to (1,0) where pickup is located, then repeats it 5 times.
        for i in range(6):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert isinstance(list(self.avatar.effects)[0], pickup_created.effects[0])
        assert list(self.avatar.effects)[0]._time_remaining == 5
        assert self.avatar.resistance == INVULNERABILITY_RESISTANCE

        # Run 5 more turns and expect the effect to expire.
        for i in range(5):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 0
        assert self.avatar.resistance == 0

    async def test_multiple_damage_boost_pickup_expiry(self, loop):
        """
        Ensure that each damage boost effect expires at the appropriate time, even if the user stacks up on them. One
        pickup will be at (1,0) and another at (3,0).
        """
        cell_one = self.game.game_state.world_map.get_cell(Location(1, 0))
        cell_two = self.game.game_state.world_map.get_cell(Location(3, 0))
        pickup_created_one = DamageBoostPickup(cell_one)
        pickup_created_two = DamageBoostPickup(cell_two)

        cell_one.interactable = pickup_created_one
        cell_two.interactable = pickup_created_two

        assert self.avatar.attack_strength == 1

        # Avatar moves EAST to (1,0) where pickup one is located.
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert isinstance(list(self.avatar.effects)[0], pickup_created_one.effects[0])
        assert len(self.avatar.effects) == 1
        assert list(self.avatar.effects)[0]._time_remaining == 10
        assert self.avatar.attack_strength == DAMAGE_BOOST_DEFAULT + 1

        # Move twice to the second pickup.
        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert isinstance(list(self.avatar.effects)[1], pickup_created_two.effects[0])
        assert len(self.avatar.effects) == 2
        assert self.avatar.attack_strength == DAMAGE_BOOST_DEFAULT * 2 + 1

        # Eight turns later, we expect the first effect to expire.
        for i in range(8):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 1
        assert list(self.avatar.effects)[0]._time_remaining == 2
        assert self.avatar.attack_strength == DAMAGE_BOOST_DEFAULT + 1

        # Two turns later, the second pickup expires too.
        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 0
        assert self.avatar.attack_strength == 1

    async def test_multiple_invulnerability_boost_pickup_expiry(self, loop):
        """
        Ensure that each invulnerability boost effect expires at the appropriate time, even if the user stacks up on
        them. One pickup will be at (1,0) and another at (3,0).
        """
        cell_one = self.game.game_state.world_map.get_cell(Location(1, 0))
        cell_two = self.game.game_state.world_map.get_cell(Location(3, 0))
        pickup_created_one = InvulnerabilityPickup(cell_one)
        pickup_created_two = InvulnerabilityPickup(cell_two)

        cell_one.interactable = pickup_created_one
        cell_two.interactable = pickup_created_two

        assert self.avatar.resistance == 0

        # Avatar moves EAST to (1,0) where pickup one is located.
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert isinstance(list(self.avatar.effects)[0], pickup_created_one.effects[0])
        assert len(self.avatar.effects) == 1
        assert list(self.avatar.effects)[0]._time_remaining == 10
        assert self.avatar.resistance == INVULNERABILITY_RESISTANCE

        # Move twice to the second pickup.
        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert isinstance(list(self.avatar.effects)[1], pickup_created_two.effects[0])
        assert len(self.avatar.effects) == 2
        assert self.avatar.resistance == INVULNERABILITY_RESISTANCE * 2

        # Eight turns later, we expect the first effect to expire.
        for i in range(8):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 1
        assert list(self.avatar.effects)[0]._time_remaining == 2
        assert self.avatar.resistance == INVULNERABILITY_RESISTANCE

        # Two turns later, the second pickup expires too.
        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert len(self.avatar.effects) == 0
        assert self.avatar.resistance == 0



from __future__ import absolute_import

import logging
import os

from django.contrib.auth.decorators import login_required
from django.core.exceptions import ValidationError
from django.http import Http404, HttpResponse, HttpResponseForbidden, JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.views.decorators.csrf import csrf_exempt, ensure_csrf_cookie
from django.views.decorators.http import require_http_methods
from django.views.generic import TemplateView
from rest_framework import mixins, status, viewsets
from rest_framework.authentication import BasicAuthentication, SessionAuthentication
from rest_framework.response import Response
from rest_framework.views import APIView

from . import forms
from . import game_renderer
from .app_settings import get_users_for_new_game
from .exceptions import UserCannotPlayGameException
from .models import Avatar, Game, LevelAttempt
from .permissions import (
    CanDeleteGameOrReadOnly,
    CsrfExemptSessionAuthentication,
    GameHasToken,
)
from .serializers import GameSerializer

LOGGER = logging.getLogger(__name__)


def _post_code_success_response(message):
    return _create_response("SUCCESS", message)


def _create_response(status, message):
    response = {"status": status, "message": message}
    return JsonResponse(response)


def _create_avatar_for_user(user, game_id, avatar_template_name="simple_avatar"):
    initial_code_file_name = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "avatar_examples/{}.py".format(avatar_template_name),
    )
    with open(initial_code_file_name) as initial_code_file:
        initial_code = initial_code_file.read()
        avatar = Avatar.objects.create(owner=user, code=initial_code, game_id=game_id)
    return avatar


@login_required
def code(request, id):
    if not request.user:
        return HttpResponseForbidden()
    game = get_object_or_404(Game, id=id)
    if not game.can_user_play(request.user):
        raise Http404
    try:
        avatar = game.avatar_set.get(owner=request.user)
    except Avatar.DoesNotExist:
        avatar = _create_avatar_for_user(request.user, id)
    if request.method == "POST":
        avatar.code = request.POST["code"]
        avatar.save()
        return HttpResponse(status=200)
    else:
        return JsonResponse({"code": avatar.code})


class GameUsersView(APIView):
    authentication_classes = (CsrfExemptSessionAuthentication, BasicAuthentication)
    permission_classes = (GameHasToken,)

    def get(self, request, id):
        game = get_object_or_404(Game, id=id)
        data = self.serialize_users(game)
        return JsonResponse(data)

    def serialize_users(self, game):
        users = {"main_avatar": None, "users": []}
        for avatar in game.avatar_set.all():
            if avatar.owner_id == game.main_user_id:
                users["main_avatar"] = avatar.id
            users["users"].append({"id": avatar.id, "code": avatar.code})
        return users


class GameViewSet(
    viewsets.GenericViewSet,
    mixins.DestroyModelMixin,
    mixins.RetrieveModelMixin,
    mixins.UpdateModelMixin,
):
    authentication_classes = (CsrfExemptSessionAuthentication, SessionAuthentication)
    queryset = Game.objects.all()
    permission_classes = (CanDeleteGameOrReadOnly,)
    serializer_class = GameSerializer

    def list(self, request):
        response = {}
        for game in Game.objects.exclude_inactive():
            serializer = GameSerializer(game)
            response[game.pk] = serializer.data
        return Response(response)


def connection_parameters(request, game_id):
    """
    An API view which returns the correct connection settings required
    to run the game in different environments. These values will change
    depending on where the project is started (ie. local, etc).
    :param request: Django request object.
    :param game_id: Integer with the ID of the game.
    :return: JsonResponse object with the contents.
    """
    env_connection_settings = game_renderer.get_environment_connection_settings(game_id)

    avatar_id, response = get_avatar_id(request, game_id)

    if avatar_id:
        env_connection_settings.update({"avatar_id": avatar_id})
        return JsonResponse(env_connection_settings)
    else:
        return response


@csrf_exempt
@require_http_methods(["POST"])
def mark_game_complete(request, id):
    game = get_object_or_404(Game, id=id)
    game.completed = True
    game.static_data = request.body
    game.save()
    return HttpResponse("Done!")


class GameTokenView(APIView):
    """
    View to Game tokens, used to prove a request comes from a game.
    """

    authentication_classes = (CsrfExemptSessionAuthentication, BasicAuthentication)
    permission_classes = (GameHasToken,)

    def get(self, request, id):
        """
        After the inital token request, we need to check where the
        request comes from. So for subsequent requests we verify that
        they came from the token-holder.
        """
        game = get_object_or_404(Game, id=id)
        self.check_object_permissions(self.request, game)
        return Response(data={"token": game.auth_token})

    def patch(self, request, id):
        game = get_object_or_404(Game, id=id)
        self.check_object_permissions(self.request, game)
        try:
            game.auth_token = request.data["token"]
            game.save()
            return Response(status=status.HTTP_200_OK)
        except KeyError:
            return Response(status=status.HTTP_403_FORBIDDEN)


class ProgramView(TemplateView):
    template_name = "players/program.html"

    def get_context_data(self, **kwargs):
        context = super(ProgramView, self).get_context_data(**kwargs)
        game = get_object_or_404(Game, id=self.kwargs["id"])
        if not game.can_user_play(self.request.user):
            raise Http404
        context["game_id"] = int(self.kwargs["id"])
        return context


@ensure_csrf_cookie
def watch_game(request, id):
    game = get_object_or_404(Game, id=id)
    if not game.can_user_play(request.user):
        raise Http404

    game.status = Game.RUNNING
    game.save()
    return game_renderer.render_game(request, game)


def watch_level(request, num):
    try:
        game = Game.objects.get(
            levelattempt__user=request.user, levelattempt__level_number=num
        )
    except Game.DoesNotExist:
        LOGGER.debug("Adding level")
        game = _add_and_return_level(num, request.user)
    LOGGER.debug("Displaying game with id %s", game.id)
    return game_renderer.render_game(request, game)


def _add_and_return_level(num, user):
    game = Game(
        generator="Level" + num, name="Level " + num, public=False, main_user=user
    )
    try:
        game.save()
    except ValidationError as e:
        LOGGER.warn(e)
        raise Http404
    game.can_play = [user]
    game.save()
    level_attempt = LevelAttempt(game=game, user=user, level_number=num)
    level_attempt.save()
    return game


@login_required
def add_game(request):
    playable_games = request.user.playable_games.all()

    if request.method == "POST":
        form = forms.AddGameForm(playable_games, data=request.POST)
        if form.is_valid():
            game = form.save(commit=False)
            game.generator = "Main"
            game.owner = request.user
            game.main_user = request.user
            game.save()
            users = get_users_for_new_game(request)
            if users is not None:
                game.can_play.add(*users)
            _create_avatar_for_user(request.user, game.id)
            return redirect("kurono/play", id=game.id)
    else:
        form = forms.AddGameForm(playable_games)
    return render(request, "players/add_game.html", {"form": form})


def current_avatar_in_game(request, game_id):
    avatar_id, response = get_avatar_id(request, game_id)

    if avatar_id:
        return JsonResponse({"current_avatar_id": avatar_id})
    else:
        return response


def get_avatar_id(request, game_id):
    avatar_id = None
    response = None

    try:
        avatar_id = game_renderer.get_avatar_id_from_user(
            user=request.user, game_id=game_id
        )
    except UserCannotPlayGameException:
        LOGGER.warning(
            "HTTP 401 returned. User {} unauthorised to play.".format(request.user.id)
        )
        response = HttpResponse("User unauthorized to play", status=401)
    except Avatar.DoesNotExist:
        LOGGER.warning(
            "Avatar does not exist for user {} in game {}".format(
                request.user.id, game_id
            )
        )
        response = HttpResponse("Avatar does not exist for this user", status=404)
    except Exception as e:
        LOGGER.error("Unknown error occurred while getting connection parameters!")
        LOGGER.error(e)
        response = HttpResponse(
            "Unknown error occurred when getting the current avatar", status=500
        )

    return avatar_id, response



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-07-03 11:51
from __future__ import unicode_literals

from django.db import migrations


def reset_tokens(apps, schema_editor):
    Game = apps.get_model("aimmo", "Game")
    for game in Game.objects.all():
        game.auth_token = ""
        game.save()


def reverse(apps, schema_editor):
    pass


class Migration(migrations.Migration):

    dependencies = [("aimmo", "0010_alter_game_token")]

    operations = [migrations.RunPython(reset_tokens, reverse_code=reverse)]



import asyncio

import pytest
from aioresponses import aioresponses
from requests import Response
from tests.test_simulation.concrete_worker import ConcreteWorker

import mock
from simulation.workers.worker import Worker

DEFAULT_RESPONSE_CONTENT = (
    b'{"action": "test_action",' b'"log": "test_log",' b'"avatar_updated": "True"}'
)

MISSING_KEY_RESPONSE_CONTENT = (
    b'{"corruptedKey": "test_action",'
    b'"log": "test_log",'
    b'"avatar_updated": "True"}'
)


def construct_test_response(status_code=200, response_content=DEFAULT_RESPONSE_CONTENT):
    post_response = Response()
    post_response.status_code = status_code
    post_response._content = response_content
    return post_response


@pytest.fixture
def worker():
    return ConcreteWorker(1, 0)


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as mocked:
        yield mocked


@pytest.mark.asyncio
async def test_fetch_data_fetches_correct_response(mock_aioresponse, worker):
    mock_aioresponse.post(
        "http://test/turn/", status=200, body=DEFAULT_RESPONSE_CONTENT
    )
    await asyncio.ensure_future(worker.fetch_data(state_view={}))

    assert worker.serialized_action == "test_action"
    assert worker.log == "test_log"
    assert worker.has_code_updated == "True"


def test_setting_defaults_works_correctly(worker):
    worker.log = "test_log_fake"
    worker.serialised_action = "test_action_fake"
    worker.has_code_updated = "test_avatar_updated_fake"

    worker._set_defaults()

    assert worker.serialized_action is None
    assert worker.log is None
    assert worker.has_code_updated is False


@pytest.mark.asyncio
async def test_fetch_data_cannot_connect_to_worker(mock_aioresponse, worker, mocker):
    mocker.patch.object(target=Worker, attribute="_set_defaults")
    mock_aioresponse.post(
        "http://test/turn/", status=500, body=MISSING_KEY_RESPONSE_CONTENT
    )
    await asyncio.ensure_future(worker.fetch_data(state_view={}))

    worker._set_defaults.assert_called_once()


@pytest.mark.asyncio
async def test_missing_key_in_worker_data(mock_aioresponse, worker, mocker):
    mocker.patch.object(target=Worker, attribute="_set_defaults")
    mock_aioresponse.post(
        "http://test/turn/", status=200, body=MISSING_KEY_RESPONSE_CONTENT
    )
    await asyncio.ensure_future(worker.fetch_data(state_view={}))

    worker._set_defaults.assert_called_once()



import asyncio

from simulation.location import Location
from simulation.interactables.pickups import InvulnerabilityPickup

from .mock_world import MockWorld


class TestInvulnerabilityPickupsAndEffects:
    def setup_method(self):
        """
        Mock a game for each test individually. MockWorld() will set up a:
        avatar manager, game state, turn manager and a map generator.
        """
        self.game = MockWorld()
        self.game.simulation_runner.add_avatar(1, Location(0, 0))
        self.cell = self.game.game_state.world_map.get_cell(Location(1, 0))

    async def test_invulnerability_pickups_increase_resistance_of_avatar(self, loop):
        """
        Avatar spawns at the origin (0,0) and should have a resistance of 0. Moves
        EAST to (1,0) and should automatically pick up the pickup and get the
        effect.
        """
        self.cell.interactable = InvulnerabilityPickup(self.cell)
        assert self.game.avatar_manager.get_avatar(1).resistance == 0
        assert self.cell.interactable.serialize() == {
            "type": "invulnerability",
            "location": {"x": self.cell.location.x, "y": self.cell.location.y},
        }

        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)
        assert self.cell.avatar.resistance == 1000
        assert len(self.cell.avatar.effects) == 1

    async def test_invulnerability_pickups_can_increase_resistance_to_2000(self, loop):
        """
        Checks if the pickup can be applied twice. First moved from ORIGIN to 1,0 ->
        then picks up the pickup, and moves to 2,0 to do the same.
        """
        self.cell.interactable = InvulnerabilityPickup(self.cell)
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )
        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)
        assert self.cell.avatar.resistance == 1000

        self.cell = self.game.game_state.world_map.get_cell(Location(2, 0))
        self.cell.interactable = InvulnerabilityPickup(self.cell)
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.cell.avatar.resistance == 2000
        assert self.cell.avatar == self.game.avatar_manager.get_avatar(1)



import unittest
import importlib
import logging

from django.test.runner import DiscoverRunner
import django.conf


LOGGER = logging.getLogger(__name__)


class DjangoAutoTestSuite(unittest.TestSuite):
    """
    This test suite configures django settings (which should be in test_settings.py), and starts a test runner.
    It allows us to run the django tests with setup.py test.
    """

    def __init__(self, *args, **kwargs):
        self._configure()
        self.test_runner = DiscoverRunner()
        tests = self.test_runner.build_suite()
        super(DjangoAutoTestSuite, self).__init__(tests=tests, *args, **kwargs)
        self.test_runner.setup_test_environment()

        self.test_dbs = self.test_runner.setup_databases()

    def _configure(self):
        test_settings = importlib.import_module('test_settings')
        setting_attrs = {attr: getattr(test_settings, attr) for attr in dir(test_settings) if '__' not in attr}

        if not django.conf.settings.configured:
            django.conf.settings.configure(**setting_attrs)

        django.setup()

    def run(self, result_obj, *args, **kwargs):
        result = super(DjangoAutoTestSuite, self).run(result_obj, *args, **kwargs)
        self.test_runner.teardown_databases(self.test_dbs)
        self.test_runner.teardown_test_environment()

        return result



from __future__ import absolute_import

from unittest import TestCase

from simulation.direction import Direction
from simulation.location import Location


class TestDirection(TestCase):
    def test_good_data(self):
        d = Direction(0, 1)
        self.assertEqual(d.x, 0)
        self.assertEqual(d.y, 1)

    def test_high_x(self):
        with self.assertRaises(ValueError):
            Direction(1.5, 0)

    def test_low_x(self):
        with self.assertRaises(ValueError):
            Direction(-1.5, 0)

    def test_high_y(self):
        with self.assertRaises(ValueError):
            Direction(0, 1.5)

    def test_low_y(self):
        with self.assertRaises(ValueError):
            Direction(0, -1.5)

    def test_too_far(self):
        with self.assertRaises(ValueError):
            Direction(1, 1)

    def test_repr(self):
        txt = repr(Direction(1, 0))
        self.assertRegex(txt, "x *= *1")
        self.assertRegex(txt, "y *= *0")

    def test_incorrect_equality(self):
        d1 = Direction(0, 1)
        l1 = Location(0, 1)
        self.assertFalse(d1 == l1)



import secrets
from base64 import b64encode as encode

import kubernetes
import pytest

from authentication import _decode_token_from_secret

TOKEN_MAX_LENGTH = 24


@pytest.fixture
def secret():
    """
    Gives a kubernetes secret object similar to the one received by the game when
    retrieving its secret. The token is randomly generated as it would be in practice,
    uses less bytes to prevent random test failure.
    """
    test_token = encode(secrets.token_urlsafe(nbytes=12).encode("utf-8"))
    return kubernetes.client.V1Secret(
        kind="Secret",
        data={"token": test_token},
        metadata=kubernetes.client.V1ObjectMeta(
            name="Test-Secret", namespace="default"
        ),
    )


def test_get_token_from_secret(secret):
    """
    Tests that we can correctly retrieve a token from the Kubernetes secret object.

    Note this is not an actual secret that would be on a cluster, it is the object returned
    when listing/reading these secrets.
    """
    token = None
    token = _decode_token_from_secret(secret)
    assert token
    assert isinstance(token, str)
    assert len(token) <= TOKEN_MAX_LENGTH



from collections import namedtuple

ReceivedAttackEvent = namedtuple(
    "ReceivedAttackEvent", ["attacking_avatar", "damage_dealt"]
)

PerformedAttackEvent = namedtuple(
    "PerformedAttackEvent", ["attacked_avatar", "target_location", "damage_dealt"]
)

FailedAttackEvent = namedtuple("FailedAttackEvent", ["target_location"])

MovedEvent = namedtuple("MovedEvent", ["source_location", "target_location"])

FailedMoveEvent = namedtuple("FailedMoveEvent", ["source_location", "target_location"])



import math
import random
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from logging import getLogger

from simulation.cell import Cell
from simulation.interactables.pickups import ALL_PICKUPS
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location

LOGGER = getLogger(__name__)

MapContext = namedtuple("MapContext", "num_avatars")


class _MapUpdater:
    __metaclass__ = ABCMeta

    @abstractmethod
    def update(self, world_map, context):
        raise NotImplementedError


class ScoreLocationUpdater(_MapUpdater):
    def update(self, world_map, context):
        for cell in world_map.score_cells():
            if random.random() < world_map.settings["SCORE_DESPAWN_CHANCE"]:
                cell.interactable.delete()

        new_num_score_locations = len(list(world_map.score_cells()))
        target_num_score_locations = int(
            math.ceil(
                context.num_avatars
                * world_map.settings["TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR"]
            )
        )
        num_score_locations_to_add = (
            target_num_score_locations - new_num_score_locations
        )
        locations = world_map._spawn_location_finder.get_random_spawn_locations(
            num_score_locations_to_add
        )
        for cell in locations:
            cell.interactable = ScoreLocation(cell)


class PickupUpdater(_MapUpdater):
    def update(self, world_map, context):
        target_num_pickups = int(
            math.ceil(
                context.num_avatars
                * world_map.settings["TARGET_NUM_PICKUPS_PER_AVATAR"]
            )
        )
        max_num_pickups_to_add = target_num_pickups - len(
            list(world_map.pickup_cells())
        )
        locations = world_map._spawn_location_finder.get_random_spawn_locations(
            max_num_pickups_to_add
        )
        for cell in locations:
            if random.random() < world_map.settings["PICKUP_SPAWN_CHANCE"]:
                cell.interactable = random.choice(ALL_PICKUPS)(cell)
                LOGGER.info(
                    "Adding new pickup at %s of type %s", cell, cell.interactable
                )


class MapExpander(_MapUpdater):
    def update(self, world_map, context):
        start_size = world_map.num_cells
        target_num_cells = int(
            math.ceil(
                context.num_avatars * world_map.settings["TARGET_NUM_CELLS_PER_AVATAR"]
            )
        )
        num_cells_to_add = target_num_cells - world_map.num_cells
        if num_cells_to_add > 0:
            self._add_outer_layer(world_map)
            assert world_map.num_cells > start_size

    def _add_outer_layer(self, world_map):
        self._add_vertical_layer(world_map, world_map.min_x() - 1)
        self._add_vertical_layer(world_map, world_map.max_x() + 1)
        self._add_horizontal_layer(world_map, world_map.min_y() - 1)
        self._add_horizontal_layer(world_map, world_map.max_y() + 1)

    def _add_vertical_layer(self, world_map, x):
        for y in range(world_map.min_y(), world_map.max_y() + 1):
            world_map.grid[Location(x, y)] = Cell(Location(x, y))

    def _add_horizontal_layer(self, world_map, y):
        for x in range(world_map.min_x(), world_map.max_x() + 1):
            world_map.grid[Location(x, y)] = Cell(Location(x, y))



import asyncio
import math

import pytest
from simulation.location import Location
from simulation.interactables.conditions import TurnState, avatar_on_cell

from .mock_world import MockWorld


@pytest.fixture
def game():
    game = MockWorld()
    game.simulation_runner.add_avatar(1, Location(0, 0))
    return game


def test_avatar_on_cell(game):
    cell = game.game_state.world_map.get_cell(Location(1, 0))
    condition = avatar_on_cell(TurnState(game.game_state.world_map, cell))
    assert not condition

    cell = game.game_state.world_map.get_cell(Location(0, 0))
    condition = avatar_on_cell(TurnState(game.game_state.world_map, cell))
    assert condition



import logging
import unittest
import psutil
import time
import kubernetes.client
from aimmo_runner import runner
from connection_api import (delete_old_database, create_custom_game_default_settings)

logging.basicConfig(level=logging.WARNING)


class TestKubernetes(unittest.TestCase):
    def setUp(self):
        """
        Sets a clean database for each test before running
        starting starting the run script again. Sleeps 60 second
        between each test to ensure stable state and loads the
        api instance from the kubernetes client.
        """
        delete_old_database()
        self.processes = runner.run(use_minikube=True, server_wait=False, capture_output=True, test_env=True)
        time.sleep(120)
        kubernetes.config.load_kube_config(context='minikube')
        self.api_instance = kubernetes.client.CoreV1Api()
        self.api_extension_instance = kubernetes.client.ExtensionsV1beta1Api()

    def tearDown(self):
        """
        Kills the process and its children peacefully.
        """
        for process in self.processes:
            try:
                parent = psutil.Process(process.pid)
            except psutil.NoSuchProcess:
                return

            children = parent.children(recursive=True)

            for child in children:
                child.terminate()

    def test_clean_starting_state_of_cluster(self):
        """
        The purpose of this test is to check the correct number
        of pods, replication controllers services, service accounts, roles/cluster roles
        and their bindings are created.
        All components created by the game will be in the "default"
        namespace.
        """

        # PODS
        api_response = self.api_instance.list_namespaced_pod("default")
        self.assertEqual(len(api_response.items), 1)
        pod_item = api_response.items[0]
        self.assertTrue(pod_item.metadata.name.startswith("aimmo-game-creator-"))
        self.assertEqual(len(pod_item.metadata.owner_references), 1)
        self.assertEqual(pod_item.metadata.owner_references[0].kind, "ReplicationController")

        # REPLICATION CONTROLLERS
        api_response = self.api_instance.list_namespaced_replication_controller("default")
        self.assertEqual(len(api_response.items), 1)
        pod_item = api_response.items[0]
        self.assertTrue(pod_item.metadata.name.startswith("aimmo-game-creator"))

        # SERVICES
        api_response = self.api_instance.list_namespaced_service("default")
        self.assertEqual(len(api_response.items), 1)
        pod_item = api_response.items[0]
        self.assertEqual(pod_item.metadata.name, "kubernetes")

        # SERVICE ACCOUNTS
        api_response = self.api_instance.list_namespaced_service_account('default')
        service_account_info = api_response.items
        single_service_account = service_account_info[1].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('game-creator' in single_service_account)

        single_service_account = service_account_info[2].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('worker' in single_service_account)

        single_service_account = service_account_info[3].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('worker-manager' in single_service_account)

        api_extension = kubernetes.client.RbacAuthorizationV1Api()

        # ROLES
        api_response = api_extension.list_namespaced_role('default')
        role_info = api_response.items[0].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('worker-manager' in role_info)

        # CLUSTER ROLES
        api_response = api_extension.list_cluster_role()
        cluster_role_info = api_response.items[0].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('game-creator' in cluster_role_info)

        # BINDINGS
        api_response = api_extension.list_namespaced_role_binding('default')
        role_binding_info = api_response.items[0].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('manage-workers' in role_binding_info)

        api_response = api_extension.list_cluster_role_binding()
        cluster_role_binding_info = api_response.items[0].metadata.annotations['kubectl.kubernetes.io/last-applied-configuration']
        self.assertTrue('create-games' in cluster_role_binding_info)

    def test_correct_initial_ingress_yaml(self):
        """
        This test will ensure that the initial yaml created on a
        fresh state of the cluster. It assumes: ingress name, no backend
        and only one specific rule, with only one path specified!
        """
        api_response = self.api_extension_instance.list_namespaced_ingress("default")
        self.assertEqual(len(api_response.items), 1)

        # NAME
        self.assertEqual(api_response.items[0].metadata.name, "aimmo-ingress")

        # NO BACKEND
        self.assertEqual(api_response.items[0].spec.backend, None)

        # RULES
        rule = api_response.items[0].spec.rules[0]
        self.assertEqual(len(api_response.items[0].spec.rules), 1)
        self.assertEqual(rule.host,
                         "local.aimmo.codeforlife.education")

        # PATHS
        path = rule.http.paths[0]
        self.assertEqual(len(rule.http.paths), 1)
        self.assertEqual(path.backend.service_name, "default-http-backend")
        self.assertEqual(path.path, None)

    @unittest.skip("Broken Test.")
    def test_adding_custom_game_sets_cluster_correctly(self):
        """
        Log into the server as admin (superuser) and create a game
        with the name "testGame", using the default settings provided.
        """

        def _wait_for_kubernetes_cluster(api_instance):
            timeout = 0

            while timeout <= 200:
                temp_response = api_instance.list_namespaced_pod("default")

                if timeout == 200:
                    self.fail("Worker not created!")

                for item in temp_response.items:
                    if item.metadata.name.startswith("aimmo-1-worker"):
                        return

                timeout += 1
                time.sleep(1)

        request_response = create_custom_game_default_settings(name="testGame")

        # WORKER
        _wait_for_kubernetes_cluster(self.api_instance)

        # SERVICE
        api_response = self.api_instance.list_namespaced_service("default")

        service_names = [service.metadata.name for service in api_response.items]
        if "game-1" not in service_names:
            self.fail("Service not created!")

        # REPLICATION CONTROLLERS
        api_response = self.api_instance.list_namespaced_replication_controller("default")

        rc_names = [rc.metadata.name for rc in api_response.items]
        if "game-1" not in rc_names:
            self.fail("Replication controller not created!")

    @unittest.skip("Not Implemented.")
    def test_adding_game_appends_path_to_ingress(self):
        """
        Adding a game (level), will append the correct path to the ingress at /game-1.
        """
        pass

    @unittest.skip("Not Implemented.")
    def test_remove_old_ingress_paths_on_startup(self):
        """
        A game is created in the minikube instance and ingress path is appended. The
        cluster is then stopped and started again with a fresh database. When this happens,
        we check that the ingress paths are returned to default again.
        """
        pass



DEFAULT_LEVEL_SETTINGS = {
    "TARGET_NUM_CELLS_PER_AVATAR": 0,
    "TARGET_NUM_SCORE_LOCATIONS_PER_AVATAR": 0,
    "SCORE_DESPAWN_CHANCE": 0,
    "TARGET_NUM_PICKUPS_PER_AVATAR": 0,
    "PICKUP_SPAWN_CHANCE": 0,
    "NO_FOG_OF_WAR_DISTANCE": 1000,
    "PARTIAL_FOG_OF_WAR_DISTANCE": 1000,
}



#!/user/bin/env python
from __future__ import absolute_import
from __future__ import print_function

import os
import platform
import socket
import time
from subprocess import CalledProcessError

import kubernetes
import yaml

from .docker_scripts import build_docker_images
from .shell_api import BASE_DIR, create_test_bin, run_command

MINIKUBE_EXECUTABLE = "minikube"
TIME_FOR_COMPONENTS_TO_DELETE = 5


def get_ip():
    """
    Get a single primary IP address. This will not return all IPs in the
    interface. See http://stackoverflow.com/a/28950776/671626
    :return: Integer with the IP of the user.
    """
    os_name = platform.system()
    if os_name == "Darwin":
        return "192.168.99.1"

    client_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    # noinspection PyBroadException
    try:
        # doesn't even have to be reachable
        client_socket.connect(("10.255.255.255", 0))
        IP = client_socket.getsockname()[0]
    except:
        IP = "127.0.0.1"
    finally:
        client_socket.close()
    return IP


def restart_ingress_addon(minikube):
    """
    Ingress needs to be restarted for old paths to be removed at startup.
    :param minikube: Executable minikube installed beforehand.
    """
    try:
        run_command([minikube, "addons", "disable", "ingress"])
    except:
        pass
    run_command([minikube, "addons", "enable", "ingress"])


def create_ingress_yaml():
    """
    Loads a ingress yaml file into a python object.
    """
    path = os.path.join(BASE_DIR, "ingress.yaml")
    with open(path) as yaml_file:
        content = yaml.safe_load(yaml_file.read())
    return content


def create_creator_yaml():
    """
    Loads a replication controller yaml file into a python object.
    """
    orig_path = os.path.join(
        BASE_DIR, "aimmo-game-creator", "rc-aimmo-game-creator.yaml"
    )
    with open(orig_path) as orig_file:
        content = yaml.safe_load(
            orig_file.read()
            .replace("latest", "test")
            .replace("REPLACE_ME", "http://%s:8000/kurono/api/games/" % get_ip())
        )
    return content


def start_cluster(minikube):
    """
    Starts the cluster unless it has been already started by the user.
    :param minikube: Executable minikube installed beforehand.
    """
    try:
        run_command([minikube, "status"], True)
        print("Cluster already running")
    except CalledProcessError:
        run_command([minikube, "start", "--memory=2048", "--cpus=2"])


def delete_components(api_instance, extensions_api_instance):
    for rc in api_instance.list_namespaced_replication_controller("default").items:
        api_instance.delete_namespaced_replication_controller(
            body=kubernetes.client.V1DeleteOptions(),
            name=rc.metadata.name,
            namespace="default",
            grace_period_seconds=0,
        )
    for pod in api_instance.list_namespaced_pod("default").items:
        api_instance.delete_namespaced_pod(
            body=kubernetes.client.V1DeleteOptions(),
            name=pod.metadata.name,
            namespace="default",
            grace_period_seconds=0,
        )
    for service in api_instance.list_namespaced_service("default").items:
        api_instance.delete_namespaced_service(
            name=service.metadata.name, namespace="default"
        )
    for ingress in extensions_api_instance.list_namespaced_ingress("default").items:
        extensions_api_instance.delete_namespaced_ingress(
            name=ingress.metadata.name,
            namespace="default",
            body=kubernetes.client.V1DeleteOptions(),
        )


def restart_pods(game_creator_yaml, ingress_yaml):
    """
    Disables all the components running in the cluster and starts them again
    with fresh updated state.
    :param game_creator_yaml: Replication controller yaml settings file.
    :param ingress_yaml: Ingress yaml settings file.
    """
    print("Restarting pods")
    kubernetes.config.load_kube_config(context="minikube")
    api_instance = kubernetes.client.CoreV1Api()
    extensions_api_instance = kubernetes.client.ExtensionsV1beta1Api()

    delete_components(api_instance, extensions_api_instance)
    time.sleep(TIME_FOR_COMPONENTS_TO_DELETE)

    extensions_api_instance.create_namespaced_ingress("default", ingress_yaml)
    api_instance.create_namespaced_replication_controller(
        body=game_creator_yaml, namespace="default"
    )


def create_roles():
    """
    Applies the service accounts, roles, and bindings for restricting
    the rights of certain pods and their processses.
    """
    run_command(["kubectl", "apply", "-Rf", "rbac"])


def start(build_target=None):
    """
    The entry point to the minikube class. Sends calls appropriately to set
    up minikube.
    """
    if platform.machine().lower() not in ("amd64", "x86_64"):
        raise ValueError("Requires 64-bit")
    create_test_bin()
    os.environ["MINIKUBE_PATH"] = MINIKUBE_EXECUTABLE
    start_cluster(MINIKUBE_EXECUTABLE)
    create_roles()
    build_docker_images(MINIKUBE_EXECUTABLE, build_target=build_target)
    restart_ingress_addon(MINIKUBE_EXECUTABLE)
    ingress = create_ingress_yaml()
    game_creator = create_creator_yaml()
    restart_pods(game_creator, ingress)
    print("Cluster ready")



import asyncio
import time

import pytest

from activity_monitor import ActivityMonitor, Timer
from asynctest import CoroutineMock

from .test_simulation.mock_communicator import MockCommunicator


@pytest.fixture
def activity_monitor():
    activity_monitor = ActivityMonitor(MockCommunicator())
    activity_monitor.callback = CoroutineMock()
    return activity_monitor


@pytest.fixture
def event_loop():
    return asyncio.get_event_loop()


@pytest.mark.asyncio
async def test_timer_stops_correctly(activity_monitor, event_loop):
    assert activity_monitor.active_users == 0
    assert not activity_monitor.timer.cancelled()

    activity_monitor.active_users = 1
    await asyncio.sleep(0.00001)

    assert activity_monitor.active_users == 1
    assert activity_monitor.timer.cancelled()


@pytest.mark.asyncio
async def test_timer_starts_correctly(activity_monitor, event_loop):
    activity_monitor.active_users = 2
    await asyncio.sleep(0.00001)

    assert activity_monitor.active_users == 2
    assert activity_monitor.timer.cancelled()

    activity_monitor.active_users = 0
    await asyncio.sleep(0.00001)

    assert activity_monitor.active_users == 0
    assert not activity_monitor.timer.cancelled()


@pytest.mark.asyncio
async def test_callback_called_when_timer_expires(activity_monitor):
    assert activity_monitor.active_users == 0
    assert not activity_monitor.timer.cancelled()

    activity_monitor.timer = Timer(timeout=0, callback=activity_monitor.callback)
    await asyncio.sleep(0.00001)

    activity_monitor.callback.assert_called()



import argparse
import os
import uuid

from django.core.management import BaseCommand
from django.test import RequestFactory
from importlib import import_module
from django.conf import settings
from django.contrib.auth.models import User
from aimmo.views import code as code_view


def _nth_dirname(path, n):
    for _ in xrange(n):
        path = os.path.dirname(path)
    return path


_PLAYERS_DIRECTORY = _nth_dirname(__file__, 3)
_AVATAR_CODES_DIRECTORY = os.path.join(_PLAYERS_DIRECTORY, "avatar_examples")

# Code file listing


def _strip_prefix(prefix, string):
    if string.startswith(prefix):
        return string[len(prefix) :]


def _get_available_code_files(base_directory):
    for dirpath, dirnames, filenames in os.walk(base_directory):
        for f in filenames:
            if not f.startswith("_") and f.endswith(".py"):
                parent_dir = _strip_prefix(base_directory, dirpath)
                yield os.path.join(parent_dir, f)


_AVATAR_CODES = list(_get_available_code_files(_AVATAR_CODES_DIRECTORY))


# Code file loading
def _load_code_file(filename):
    if not filename.endswith(".py"):
        filename = filename + ".py"

    filepath = os.path.join(_AVATAR_CODES_DIRECTORY, filename)

    with open(filepath) as f:
        return f.read()


class LoadCodeAction(argparse.Action):
    def __init__(self, option_strings, dest, **kwargs):
        super(LoadCodeAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string):
        values = _load_code_file(values)
        setattr(namespace, self.dest, values)


class Command(BaseCommand):
    # Show this when the user types help
    help = "Generate users for the game"

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.engine = import_module(settings.SESSION_ENGINE)
        self.request_factory = RequestFactory()

    def add_arguments(self, parser):
        parser.add_argument("num-users", type=int, help="Number of users to create")
        parser.add_argument(
            "avatar-code",
            choices=_AVATAR_CODES,
            action=LoadCodeAction,
            help="The code to use for the avatar.",
        )
        parser.add_argument("game-id", type=int)

    # A command must define handle()
    def handle(self, *args, **options):
        num_users = options["num-users"]
        code = options["avatar-code"]
        game_id = options["game-id"]

        for _ in xrange(num_users):
            random_string = str(uuid.uuid4())[:8]
            username = "zombie-%s" % random_string
            password = "123"
            user = self.create_user(username, password)
            self.post_code(user, code, game_id)

    def create_user(self, username, password):
        user = User.objects.create_user(username, "user@example.com", password)
        self.stdout.write("Created user %s with password: %s" % (username, password))
        return user

    def post_code(self, user, player_code, game_id):
        request = self.request_factory.post("/any_path", data={"code": player_code})
        session_key = None
        request.session = self.engine.SessionStore(session_key)
        request.user = user

        response = code_view(request, game_id)
        if response.status_code == 200:
            self.stdout.write("Posted code for player %s" % user)
        else:
            raise Exception("Failed to submit code for player %s" % user)



from unittest import TestCase

import service
from simulation.avatar.avatar_manager import AvatarManager
from simulation.game_state import GameState
from simulation.interactables.score_location import ScoreLocation
from simulation.location import Location
from simulation.world_map import WorldMap

from .test_simulation.dummy_avatar import MoveEastDummy
from .test_simulation.maps import MockCell, MockPickup


class TestService(TestCase):
    class DummyAvatarManager(AvatarManager):
        avatars = [MoveEastDummy(1, Location(0, -1))]

    @classmethod
    def setUpClass(cls):
        """ Register the api endpoints """
        cls.game_api = service.GameAPI(worker_manager=None, game_state=None)

    def setUp(self):
        """
        Sets up the JSON of the world state generated by the service file for testing.
        """
        self.avatar_manager = self.DummyAvatarManager()

        CELLS = [
            [
                {
                    "interactable": MockPickup("b"),
                    "avatar": self.avatar_manager.avatars[0],
                },
                {},
                {},
            ],
            [{}, {"habitable": False}, {"interactable": MockPickup("a")}],
        ]

        grid = {
            Location(x, y - 1): MockCell(Location(x, y - 1), **CELLS[x][y])
            for y in range(3)
            for x in range(2)
        }
        grid[Location(0, 1)].interactable = ScoreLocation(grid[Location(0, 1)])
        test_game_state = GameState(WorldMap(grid, {}), self.avatar_manager)
        self.world_state_json = test_game_state.serialize()

    def test_correct_json_player_dictionary(self):
        """
        Ensures the "players" element of the get_game_state() JSON returns the correct information for the dummy
        avatar provided into the world.

        NOTE: Orientation (and others) may be hard coded. This test WILL and SHOULD fail if the functionality is added.
        """
        player_list = self.world_state_json["players"]
        self.assertEqual(len(player_list), 1)
        details = player_list[0]
        print("Player list: {}".format(player_list))
        self.assertEqual(details["id"], 1)
        self.assertEqual(details["location"]["x"], 0)
        self.assertEqual(details["location"]["y"], -1)
        self.assertEqual(details["health"], 5)
        self.assertEqual(details["orientation"], "north")
        self.assertEqual(details["score"], 0)

    def test_correct_json_score_locations(self):
        """
        Ensures the correct score location in the "score_locations" element; is returned by the JSON.
        """
        interactable_list = self.world_state_json["interactables"]
        for interactable in interactable_list:
            if "ScoreLocation" in interactable:
                self.assertEqual(interactable["location"]["x"], 0)
                self.assertEqual(interactable["location"]["y"], 1)

    def test_correct_json_north_east_corner(self):
        """
        Top right corner of the map must be correct to determine the map size.
        """
        north_east_corner = self.world_state_json["northEastCorner"]
        self.assertEqual(north_east_corner["x"], 1)
        self.assertEqual(north_east_corner["y"], 1)

    def test_correct_json_south_west_corner(self):
        """
        Bottom left corner of the map must be correct to determine the map size.
        """
        south_west_corner = self.world_state_json["southWestCorner"]
        self.assertEqual(south_west_corner["x"], 0)
        self.assertEqual(south_west_corner["y"], -1)

    def test_correct_json_era(self):
        """
        Ensure that the era (for the assets in Unity) is correct.

        NOTE: This is hard coded right now to "less_flat". This test should fail when this functionality is added.
        """
        era = self.world_state_json["era"]
        self.assertEqual(era, "less_flat")

    def test_correct_json_world_interactables_returned_is_correct_amount(self):
        """
        The JSON returns the correct amount of pickups.
        """
        interactable_list = self.world_state_json["interactables"]
        self.assertEqual(len(interactable_list), 0)

    def test_correct_json_world_obstacles(self):
        """
        JSON generated must return correct location, width, height, type and orientation about obstacles.

        NOTE: Obstacles are highly hard coded right now. Only location changes. If any functionality is added, this test
              WILL and SHOULD fail.
        """
        obstacle_list = self.world_state_json["obstacles"]
        self.assertEqual(len(obstacle_list), 1)
        self.assertEqual(obstacle_list[0]["location"]["x"], 1)
        self.assertEqual(obstacle_list[0]["location"]["y"], 0)
        self.assertEqual(obstacle_list[0]["orientation"], "north")
        self.assertEqual(obstacle_list[0]["width"], 1)
        self.assertEqual(obstacle_list[0]["height"], 1)
        self.assertEqual(obstacle_list[0]["type"], "wall")



from django.conf.urls import include, url
from django.contrib.auth import views as auth_views
from django.contrib.auth.decorators import login_required
from django.views.generic import RedirectView, TemplateView
from django_js_reverse.views import urls_js
from rest_framework import routers

from aimmo import views

HOMEPAGE_REGEX = r"^kurono/"


router = routers.SimpleRouter()
router.register(r"games", views.GameViewSet)


urlpatterns = [
    url(
        r"^$",
        login_required(TemplateView.as_view(template_name="players/home.html")),
        name="kurono/home",
    ),
    url(r"^accounts/login/$", auth_views.login, name="kurono/login"),
    url(
        r"^accounts/logout/$",
        auth_views.logout,
        {"next_page": "kurono/logout_success"},
        name="kurono/logout",
    ),
    url(
        r"^accounts/logout_success/$",
        TemplateView.as_view(template_name="registration/success_logout.html"),
        name="kurono/logout_success",
    ),
    url(
        r"^play/(?P<id>[0-9]+)/$", login_required(views.watch_game), name="kurono/play"
    ),
    url(
        r"^statistics/$",
        TemplateView.as_view(template_name="players/statistics.html"),
        name="kurono/statistics",
    ),
    url(r"^api/code/(?P<id>[0-9]+)/$", views.code, name="kurono/code"),
    url(
        r"^api/games/(?P<id>[0-9]+)/users/$",
        views.GameUsersView.as_view(),
        name="kurono/game_user_details",
    ),
    url(
        r"^api/games/(?P<id>[0-9]+)/token/$",
        views.GameTokenView.as_view(),
        name="kurono/game_token",
    ),
    url(
        r"^api/games/(?P<game_id>[0-9]+)/connection_parameters/$",
        views.connection_parameters,
        name="kurono/connection_parameters",
    ),
    url(
        r"^api/games/(?P<id>[0-9]+)/complete/$",
        views.mark_game_complete,
        name="kurono/complete_game",
    ),
    url(
        r"^api/games/(?P<game_id>[0-9]+)/current_avatar/$",
        views.current_avatar_in_game,
        name="kurono/current_avatar_in_game",
    ),
    url(r"^api/", include(router.urls)),
    url(
        r"^jsreverse/$", urls_js, name="kurono/js_reverse"
    ),  # TODO: Pull request to make django_js_reverse.urls
    url(r"^games/new/$", views.add_game, name="kurono/new_game"),
    # TODO: this is a quickfix for redirecting for the Unity resources
    url(
        r"^watch/(?P<id>[0-9]+)/(?P<resource>.[0-9A-Za-z/.]+)$",
        RedirectView.as_view(url="/static/unity/%(resource)s", permanent=False),
    ),
    url(
        r"^favicon/.ico$",
        RedirectView.as_view(url="/static/favicon.ico", permanent=True),
    ),
]



from simulation.action import MoveAction


class Cell(object):
    """
    Any position on the world grid.
    """

    def __init__(self, location, habitable=True, partially_fogged=False):
        self.location = location
        self.habitable = habitable
        self.avatar = None
        self.interactable = None
        self.partially_fogged = partially_fogged
        self.actions = []

    def __repr__(self):
        return "Cell({} h={} a={} i={} f{})".format(
            self.location,
            self.habitable,
            self.avatar,
            self.interactable,
            self.partially_fogged,
        )

    def __eq__(self, other):
        return self.location == other.location

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.location)

    @property
    def moves(self):
        return [move for move in self.actions if isinstance(move, MoveAction)]

    @property
    def is_occupied(self):
        return self.avatar is not None

    def serialize(self):
        if self.partially_fogged:
            return {
                "location": self.location.serialize(),
                "partially_fogged": self.partially_fogged,
            }
        else:
            return {
                "avatar": self.avatar.serialize() if self.avatar else None,
                "habitable": self.habitable,
                "location": self.location.serialize(),
                "interactable": self.interactable.serialize()
                if self.interactable
                else None,
                "partially_fogged": self.partially_fogged,
            }



import asyncio

from tests.test_simulation.dummy_avatar import (
    DeadDummy,
    MoveEastDummy,
    MoveNorthDummy,
    MoveSouthDummy,
    MoveWestDummy,
    WaitDummy,
)

from simulation import map_generator
from simulation.location import Location
from simulation.simulation_runner import ConcurrentSimulationRunner

from .mock_world import MockWorld


class TestMovementsInMap:

    SETTINGS = {"START_HEIGHT": 50, "START_WIDTH": 50, "OBSTACLE_RATIO": 0}

    def set_up_environment(
        self,
        dummy_list=None,
        location=Location(0, 0),
        map_generator_class=map_generator.Main,
    ):
        """
        Utility method for testing.
        """
        self.game = MockWorld(
            TestMovementsInMap.SETTINGS,
            dummy_list,
            map_generator_class,
            ConcurrentSimulationRunner,
        )
        self.game.simulation_runner.add_avatar(player_id=1, location=location)
        self.avatar = self.game.avatar_manager.get_avatar(1)

    async def set_up_and_make_movements_in_a_single_direction(
        self, dummy_list, number_of_movements, spawn=Location(0, 0)
    ):
        """
        Template function for repetitive movements in a single direction.
        """
        self.set_up_environment(dummy_list, spawn)
        assert self.avatar.location == spawn

        for i in range(number_of_movements):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

    async def test_movement_five_times_in_all_directions(self, loop):
        """
        Moves the avatar to the edge of the map. Each time it moves the avatar 5 times
        from origin in all cardinal directions.
        """

        # East.
        await self.set_up_and_make_movements_in_a_single_direction([MoveEastDummy], 5)
        assert self.avatar.location == Location(5, 0)

        # West.
        await self.set_up_and_make_movements_in_a_single_direction([MoveWestDummy], 5)
        assert self.avatar.location == Location(-5, 0)

        # North.
        await self.set_up_and_make_movements_in_a_single_direction([MoveNorthDummy], 5)
        assert self.avatar.location == Location(0, 5)

        # South.
        await self.set_up_and_make_movements_in_a_single_direction([MoveSouthDummy], 5)
        assert self.avatar.location == Location(0, -5)

    async def test_move_towards_map_boundaries(self, loop):
        """
        Tests game behaviour when the avatar tries to move towards one of the four of the
        maps boundaries.
        """

        # North boundary.
        await self.set_up_and_make_movements_in_a_single_direction(
            [MoveNorthDummy], 2, Location(0, 25)
        )

        assert not self.game.game_state.world_map.is_on_map(Location(0, 26))
        assert self.avatar.location == Location(0, 25)

        # South boundary.
        await self.set_up_and_make_movements_in_a_single_direction(
            [MoveSouthDummy], 2, Location(0, -24)
        )

        assert not self.game.game_state.world_map.is_on_map(Location(0, -25))
        assert self.avatar.location == Location(0, -24)

        # East boundary.
        await self.set_up_and_make_movements_in_a_single_direction(
            [MoveEastDummy], 2, Location(25, 0)
        )

        assert not self.game.game_state.world_map.is_on_map(Location(26, 0))
        assert self.avatar.location == Location(25, 0)

        # West boundary.
        await self.set_up_and_make_movements_in_a_single_direction(
            [MoveWestDummy], 2, Location(-24, 0)
        )

        assert not self.game.game_state.world_map.is_on_map(Location(-25, 0))
        assert self.avatar.location == Location(-24, 0)

    async def test_avatar_cannot_move_into_obstacle(self, loop):
        """
        Make sure that an avatar will stay in its location when trying to move into a
        obstacle cell.
        """
        self.set_up_environment([MoveEastDummy])
        obstacle_cell = self.game.game_state.world_map.get_cell(Location(2, 0))
        obstacle_cell.habitable = False
        assert self.avatar.location == Location(0, 0)

        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert self.avatar.location == Location(1, 0)

    async def test_avatars_cannot_go_into_each_other(self, loop):
        """
        Two avatars moving in the same direction towards each other.
        """
        # Even number of cells between two avatars.
        self.set_up_environment([MoveEastDummy, MoveWestDummy])
        self.game.simulation_runner.add_avatar(2, Location(3, 0))
        avatar_two = self.game.avatar_manager.get_avatar(2)

        assert self.avatar.location == Location(0, 0)
        assert avatar_two.location == Location(3, 0)
        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        # Avatar 1 & Avatar 2 only managed to move once.
        assert self.avatar.location == Location(1, 0)
        assert avatar_two.location == Location(2, 0)

        # Odd number of cells between two avatars.
        self.set_up_environment([MoveEastDummy, MoveWestDummy])
        self.game.simulation_runner.add_avatar(2, Location(4, 0))
        avatar_two = self.game.avatar_manager.get_avatar(2)

        assert self.avatar.location == Location(0, 0)
        assert avatar_two.location == Location(4, 0)

        for i in range(2):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        # Avatar 1 & Avatar 2 managed to only move only once.
        assert self.avatar.location == Location(1, 0)
        assert avatar_two.location == Location(3, 0)

        # Live avatar can't move into a square occupied by a 'dead' (no worker) avatar
        self.set_up_environment([DeadDummy, MoveWestDummy])
        self.game.simulation_runner.add_avatar(2, Location(1, 0))
        avatar_two = self.game.avatar_manager.get_avatar(2)

        assert self.avatar.location == Location(0, 0)
        assert avatar_two.location == Location(1, 0)
        await self.game.simulation_runner.run_single_turn(
            self.game.avatar_manager.get_player_id_to_serialized_action()
        )

        assert self.avatar.location == Location(0, 0)
        assert avatar_two.location == Location(1, 0)

    async def test_wait_action_on_a_single_avatar(self, loop):
        """
        Ensures a returned WaitAction will keep the avatar in its initial location.
        """
        self.set_up_environment(dummy_list=[WaitDummy])
        assert self.avatar.location == Location(0, 0)

        for i in range(5):
            await self.game.simulation_runner.run_single_turn(
                self.game.avatar_manager.get_player_id_to_serialized_action()
            )

        assert self.avatar.location == Location(0, 0)



from importlib import import_module

from django.apps import apps
from django.db import connection
from django.test import TestCase

from aimmo.models import Game

# Our filename starts with a number, so we use import_module
data_migration = import_module("aimmo.migrations.0011_reset_game_tokens")


class ResetGameTokensTests(TestCase):
    def __init__(self, *args, **kwargs):
        super(ResetGameTokensTests, self).__init__(*args, **kwargs)

        self.token = "I AM A TOKEN!"

    def test_reset_token(self):
        game = Game.objects.create(id=1, name="test", public=True)
        game.auth_token = self.token
        game.save()

        data_migration.reset_tokens(apps, connection.schema_editor())

        game = Game.objects.get(id=1)
        self.assertEqual(game.auth_token, "")



# -*- coding: utf-8 -*-
import re
import sys

from setuptools import find_packages, setup

with open('aimmo/__init__.py', 'r') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(),
        re.MULTILINE
    ).group(1)

try:
    from semantic_release import setup_hook

    setup_hook(sys.argv)
except ImportError:
    pass

setup(
    name="aimmo",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django >= 1.10.8, <= 1.11.24",
        "django-autoconfig >= 0.3.6, < 1.0.0",
        "django-js-reverse",
        "djangorestframework>=3.8.2, < 3.9.0",
        "eventlet",
        "requests",
        "six",
        "hypothesis",
        "psutil >= 5.4, < 5.5",
    ],
    tests_require=[
        "httmock",
        "mock == 2.0.0",
        "docker >= 3.5, < 3.6",
        "kubernetes == 5.0.0",
        "PyYAML == 4.2b1",
    ],
    version=version,
    zip_safe=False,
)



from django.contrib.auth.models import AnonymousUser, User
from django.test import TestCase

from aimmo import models, app_settings

app_settings.GAME_SERVER_URL_FUNCTION = lambda game_id: (
    "base %s" % game_id,
    "path %s" % game_id,
)
app_settings.GAME_SERVER_PORT_FUNCTION = lambda game_id: 0
app_settings.GAME_SERVER_SSL_FLAG = True


class TestModels(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.user1 = User.objects.create_user("test", "test@example.com", "password")
        cls.user1.save()
        cls.user2 = User.objects.create_user("test2", "test2@example.com", "password")
        cls.user2.save()
        cls.game = models.Game(id=1, name="test", public=False)
        cls.game.save()

    def setUp(self):
        self.game.refresh_from_db()

    def _make_game_private_for_user(self, user):
        self.game.public = False
        self.game.can_play = [user]

    def _run_test_authorisation_to_play(self, user, can_play):
        self._make_game_private_for_user(self.user1)
        self.assertEqual(self.game.can_user_play(user), can_play)

    def test_public_games_can_be_accessed(self):
        self.game.public = True
        self.assertTrue(self.game.can_user_play(self.user1))
        self.assertTrue(self.game.can_user_play(self.user2))

    def test_anon_user_can_play_public_game(self):
        self.game.public = True
        self.assertTrue(self.game.can_user_play(AnonymousUser()))

    def test_authorised_user_can_play(self):
        self._run_test_authorisation_to_play(self.user1, True)

    def test_unauthorised_user_cannot_play(self):
        self._run_test_authorisation_to_play(self.user2, False)

    def test_game_active_by_default(self):
        self.assertTrue(self.game.is_active)

    def test_completed_game_inactive(self):
        self.game.completed = True
        self.game.save()
        self.assertFalse(self.game.is_active)



import abc
import heapq
import logging
import random
from itertools import tee
from queue import PriorityQueue
from typing import Any

from six.moves import range, zip

from simulation.direction import ALL_DIRECTIONS
from simulation.game_state import GameState
from simulation.interactables.score_location import ScoreLocation
from simulation.level_settings import DEFAULT_LEVEL_SETTINGS
from simulation.location import Location
from simulation.world_map import WorldMap, WorldMapStaticSpawnDecorator

LOGGER = logging.getLogger(__name__)


class _BaseGenerator(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, settings):
        self.settings = settings

    def get_game_state(self, avatar_manager):
        return GameState(self.get_map(), avatar_manager, self.check_complete)

    def check_complete(self, game_state):
        return False

    @abc.abstractmethod
    def get_map(self):
        pass


class Main(_BaseGenerator):
    def get_map(self):
        height = self.settings["START_HEIGHT"]
        width = self.settings["START_WIDTH"]
        world_map = WorldMap.generate_empty_map(height, width, self.settings)

        # We set one non-corner edge cell as empty, to ensure that the map can be expanded
        always_empty_edge_x, always_empty_edge_y = get_random_edge_index(world_map)
        always_empty_location = Location(always_empty_edge_x, always_empty_edge_y)

        for cell in shuffled(world_map.all_cells()):
            if (
                cell.location != always_empty_location
                and random.random() < self.settings["OBSTACLE_RATIO"]
            ):
                cell.habitable = False
                # So long as all habitable neighbours can still reach each other, then the
                # map cannot get bisected.
                if not _all_habitable_neighbours_can_reach_each_other(cell, world_map):
                    cell.habitable = True

        return world_map


def _get_edge_coordinates(height, width):
    for x in range(width):
        for y in range(height):
            yield x, y


def shuffled(iterable):
    values = list(iterable)
    random.shuffle(values)
    return iter(values)


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def _all_habitable_neighbours_can_reach_each_other(cell, world_map):
    neighbours = get_adjacent_habitable_cells(cell, world_map)

    assert len(neighbours) >= 1
    neighbour_pairs = ((n1, n2) for n1, n2 in pairwise(neighbours))
    shortest_path_exists = (
        get_shortest_path_between(n1, n2, world_map) is not None
        for n1, n2 in neighbour_pairs
    )
    return all(shortest_path_exists)


def get_shortest_path_between(source_cell, destination_cell, world_map):
    def manhattan_distance_to_destination_cell(this_branch):
        branch_tip_location = this_branch[-1].location
        x_distance = abs(branch_tip_location.x - destination_cell.location.x)
        y_distance = abs(branch_tip_location.y - destination_cell.location.y)
        return x_distance + y_distance + len(this_branch)

    branches = PriorityQueuef(
        key=manhattan_distance_to_destination_cell, init_items=[[source_cell]]
    )
    visited_cells = set()

    while not branches.queue.empty():
        branch = branches.pop()

        for cell in get_adjacent_habitable_cells(branch[-1], world_map):
            if cell in visited_cells:
                continue

            visited_cells.add(cell)

            new_branch = branch + [cell]

            if cell == destination_cell:
                return new_branch

            branches.push(new_branch)

    return None


def get_random_edge_index(world_map, rng=random):
    num_row_cells = world_map.num_rows - 2
    num_col_cells = world_map.num_cols - 2
    num_edge_cells = 2 * num_row_cells + 2 * num_col_cells
    random_cell = rng.randint(0, num_edge_cells - 1)

    if 0 <= random_cell < num_col_cells:
        # random non-corner cell on the first row
        return random_cell + 1 + world_map.min_x(), world_map.min_y()
    random_cell -= num_col_cells

    if 0 <= random_cell < num_col_cells:
        # random non-corner cell on the last row
        return random_cell + 1 + world_map.min_x(), world_map.max_y()
    random_cell -= num_col_cells

    if 0 <= random_cell < num_row_cells:
        # random non-corner cell on the first column
        return world_map.min_x(), world_map.min_y() + random_cell + 1
    random_cell -= num_row_cells

    assert 0 <= random_cell < num_row_cells
    # random non-corner cell on the last column
    return world_map.max_x(), world_map.min_y() + random_cell + 1


def get_adjacent_habitable_cells(cell, world_map):
    adjacent_locations = [cell.location + d for d in ALL_DIRECTIONS]
    adjacent_locations = [
        location for location in adjacent_locations if world_map.is_on_map(location)
    ]

    adjacent_cells = [world_map.get_cell(location) for location in adjacent_locations]
    return [c for c in adjacent_cells if c.habitable]


class PriorityEntry(object):
    priority: int
    value: Any

    def __init__(self, priority, value):
        self.priority = priority
        self.value = value

    def __lt__(self, other):
        return self.priority < other.priority


class PriorityQueuef(object):
    def __init__(self, key, init_items=tuple()):
        self.key = key
        self.heap = [self._build_tuple(i) for i in init_items]
        self.queue = PriorityQueue()
        for item in self.heap:
            self.queue.put(item, block=True)

    def _build_tuple(self, item):
        return PriorityEntry(self.key(item), item)

    def push(self, item):
        to_push = self._build_tuple(item)
        self.queue.put(to_push, block=True)

    def pop(self):
        entry: PriorityEntry = self.queue.get(block=True)
        return entry.value

    def __len__(self):
        return self.queue.qsize()


class _BaseLevelGenerator(_BaseGenerator):
    __metaclass__ = abc.ABCMeta

    def __init__(self, *args, **kwargs):
        super(_BaseLevelGenerator, self).__init__(*args, **kwargs)
        self.settings.update(DEFAULT_LEVEL_SETTINGS)


class Level1(_BaseLevelGenerator):
    def get_map(self):
        world_map = WorldMap.generate_empty_map(1, 5, self.settings)
        world_map = WorldMapStaticSpawnDecorator(world_map, Location(-2, 0))
        world_map.get_cell(Location(2, 0)).interactable = ScoreLocation(
            world_map.get_cell(Location(2, 0))
        )
        return world_map

    def check_complete(self, game_state):
        try:
            main_avatar = game_state.get_main_avatar()
        except KeyError:
            return False
        return main_avatar.score > 0



import os
dirname = os.path.dirname
abspath = os.path.abspath

db_name = os.path.join(abspath(dirname(dirname(abspath(__file__)))), 'db.sqlite3')

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': db_name,
    },
}

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'aimmo',
    'integration_tests',
    'aimmo_runner',
]

PIPELINE_ENABLED = False
ROOT_URLCONF = 'django_autoconfig.autourlconf'
STATIC_ROOT = '.tests_static/'
SECRET_KEY = 'test_key'
GAME_SERVER_URL_FUNCTION = lambda game_id: ('base %s' % game_id, 'path %s' % game_id)
GAME_SERVER_PORT_FUNCTION = lambda game_id: 8001
GAME_SERVER_SSL_FLAG = False
from django_autoconfig.autoconfig import configure_settings  # noqa: E402
configure_settings(globals())


# -*- coding: utf-8 -*-
from setuptools import find_packages, setup


setup(
    name='integration-tests',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'requests',
    ],
    tests_require=[
        'requests',
        'psutil',
        'mock',
        'kubernetes == 5.0.0',
    ],
    test_suite='test_utils.test_suite.DjangoAutoTestSuite',
    zip_safe=False,
)



from __future__ import absolute_import

from unittest import TestCase

from simulation.game_state import GameState
from simulation.location import Location

from .dummy_avatar import DummyAvatar, DummyAvatarManager
from .maps import AvatarMap, EmptyMap, InfiniteMap


class TestGameState(TestCase):
    def game_state_with_two_avatars(self, world_map=None, avatar_manager=None):
        if world_map is None:
            world_map = EmptyMap()
        if avatar_manager is None:
            avatar_manager = DummyAvatarManager()

        avatar = DummyAvatar(1, (0, 0))
        other_avatar = DummyAvatar(2, (0, 0))
        other_avatar.marked = True
        avatar_manager.avatars_by_id[1] = avatar
        avatar_manager.avatars_by_id[2] = other_avatar
        game_state = GameState(world_map, avatar_manager)

        return (game_state, avatar, world_map, avatar_manager)

    def test_no_main_avatar_by_default(self):
        state = GameState(EmptyMap(), DummyAvatarManager())
        with self.assertRaises(KeyError):
            state.get_main_avatar()

    def test_get_main_avatar(self):
        (game_state, avatar, _, _) = self.game_state_with_two_avatars()
        game_state.main_avatar_id = avatar.player_id
        self.assertEqual(game_state.get_main_avatar(), avatar)



from __future__ import absolute_import

import abc
from unittest import TestCase

from simulation.interactables import effects

from .dummy_avatar import DummyAvatar
from .maps import MockCell


class _BaseCases(object):
    class BaseEffectTestCase(TestCase):
        __metaclass__ = abc.ABCMeta

        @abc.abstractmethod
        def make_effect(self, avatar):
            pass

        def setUp(self):
            self.cell = MockCell()
            self.avatar = DummyAvatar(1, None)
            self.effect = self.make_effect(self.avatar)
            self.avatar.effects.add(self.effect)

        def assertNoEffects(self):
            self.assertEqual(len(list(self.avatar.effects)), 0)

        def test_effect_removed(self):
            self.effect.remove()
            self.assertNoEffects()

        def test_effect_expires(self):
            for _ in range(10):
                self.effect.on_turn()
            self.assertTrue(self.effect.is_expired)


class TestInvulnerabilityEffect(_BaseCases.BaseEffectTestCase):
    def make_effect(self, *args):
        return effects.InvulnerabilityEffect(*args)

    def test_resistance_increases(self):
        self.assertEqual(self.avatar.resistance, 1000)

    def test_resistance_decreases(self):
        self.effect.remove()
        self.assertEqual(self.avatar.resistance, 0)

    def test_resistance_cannot_be_removed_twice(self):
        self.effect.remove()
        self.assertRaises(KeyError, self.effect.remove)


class TestDamageBoostEffect(_BaseCases.BaseEffectTestCase):
    def make_effect(self, *args):
        return effects.DamageBoostEffect(*args)

    def test_damage_increases(self):
        self.assertEqual(self.avatar.attack_strength, 6)

    def test_damage_decreases(self):
        self.effect.remove()
        self.assertEqual(self.avatar.attack_strength, 1)

    def test_damage_cannot_be_removed_twice(self):
        self.effect.remove()
        self.assertRaises(KeyError, self.effect.remove)
